import copy
import math
import numpy as np
import json
import requests

from plantpredict.plant_predict_entity import PlantPredictEntity
from plantpredict.error_handlers import handle_refused_connection, handle_error_response, APIError
from plantpredict.enumerations import ModuleOrientationEnum, TrackingTypeEnum, FacialityEnum


class PowerPlant(PlantPredictEntity):
    """
    Represents the hierarchical structure of a power plant in PlantPredict. There is a one-to-one relationship between a
    :py:class:`~plantpredict.powerplant.PowerPlant` and :py:class:`~plantpredict.prediction.Prediction`. It is linked to
    that prediction via the attributes :py:attr:`project_id` and :py:attr:`prediction_id`.

    All classes that inherit from :py:class:`~plantpredict.plant_predict_entity.PlantPredictEntity` follow the same
    general usage pattern. The core class methods (:py:class:`~plantpredict.powerplant.PowerPlant.get`,
    :py:class:`~plantpredict.powerplant.PowerPlant.create`, and :py:class:`~plantpredict.powerplant.PowerPlant.update`)
    require that certain attributes be assigned to the instance of the class in order to run successfully, rather than
    requiring direct variable inputs to the method call itself. For methods beyond these four, the input requirements
    might be either attribute assignments or variable inputs to the method.

    Sample code for properly building a :py:class:`~plantpredict.powerplant.PowerPlant` can be found in
    :ref:`example_usage`. While a new :py:class:`~plantpredict.powerplant.PowerPlant` can be initialized via its
    :py:meth:`~plantpredict.powerplant.PowerPlant.__init__` method, as in the following example:

    .. code-block:: python

        powerplant = plantpredict.powerplant.PowerPlant(api, project_id=1, prediction_id=2)

    it is recommended to use the :py:class:`~plantpredict.api.Api` factory method
    :py:meth:`~plantpredict.api.Api.powerplant`, as in the following example:

    .. code-block:: python

        powerplant = api.powerplant(project_id=1, prediction_id=2)

    where both cases assume that :py:data:`api` is a properly defined :py:class:`~plantpredict.api.Api` object.

    Note on parameters listed below: This list of attributes is comprehensive, but does not encompass 100% of parameters
    that might be available via :py:meth:`~plantpredict.powerplant.PowerPlant.get` after the associated prediction is
    run. The list includes all relevant attributes that a user should/can set upon building the
    :py:class:`~plantpredict.powerplant.PowerPlant`, plus some of the post-prediction-run parameters.

    :param plantpredict.api.Api api: An properly initialized instance of the PlantPredict API client class,
                                     :py:class:`~plantpredict.api.Api`, which is used for authentication with the
                                     PlantPredict servers, given a user's unique API credentials.
    :param project_id: Unique identifier for the :py:class:`~plantpredict.project.Project` with which to associate the
                       power plant. Must represent a valid, exiting project in the PlantPredict database.
    :type project_id: int, None
    :param prediction_id: Unique identifier for the :py:class:`~plantpredict.prediction.Prediction` with which to
                          associate the power plant. Must represent a valid, existing Prediction on the given Project in
                          the PlantPredict database, as represented by the input :py:data:`project_id`.
    :type prediction_id: int, None
    :param bool use_cooling_temp: If :py:data:`True`, the :py:attr:`kva_rating` of each inverter in the power plant is
                                  calculated based on the 99.6 cooling temperature of the nearest ASHRAE station to the
                                  corresponding :py:class:`~plantpredict.project.Project` (as specified by
                                  :py:attr:`project_id`), the elevation of the
                                  :py:class:`~plantpredict.project.Project`, and the elevation/temperature curves of the
                                  inverter model specified by :py:data:`inverter_id`. Defaults to :py:data:`True`. If
                                  :py:data:`False`, the :py:attr:`kva_rating` of each inverter in the power plant is set
                                  as the :py:attr:`apparent_power` of the inverter model specified by
                                  :py:data:`inverter_id`.
    :param float lgia_limitation: Maximum power output limit for power plant according to its Large Generator
                                  Interconnection Agreement (LGIA). Must be between :py:data:`0` and :py:data:`2000` -
                                  units :py:data:`[MWac]`.
    :param float availability_loss: Accounts for losses due to any plant-wide outage events such as inverter
                                    shutdowns/failures. Must be between :py:data:`0` and :py:data:`25` - units
                                    :py:data:`[%]`.
    :param float power_factor: The ratio of the power that can be used and the product of the operating current and
                               voltage (also referred to as Plant kVA Derate). Defaults to :py:data:`1.0`. Must be
                               between :py:data:`0` and :py:data:`1`, where :py:data:`1` is a "unity" power factor.
                               Defaults to :py:data:`1.0` in :py:meth:`~plantpredict.powerplant.PowerPlant.__init__` and
                               automatically recalculated when :py:meth:`~plantpredict.powerplant.PowerPlant.create`
                               called.
    :param list transformers: Defaults to an empty list (:py:data:`[]`). See "Example contents of
                              :py:attr:`transformers`" below for sample contents. Use the "power plant builder" method
                              :py:meth:`~plantpredict.powerplant.PowerPlant.add_transformer` to easily add a new
                              transformer to the attribute :py:attr:`transformers`.
    :param list transmission_lines: Defaults to an empty list (:py:data:`[]`). See "Example contents of
                                    :py:attr:`transmission_lines`" below for sample contents. Use the
                                    "power plant builder" method
                                    :py:meth:`~plantpredict.powerplant.PowerPlant.add_transmission_line` to easily add a
                                    new transmission line to the attribute :py:attr:`transmission_lines`.
    :param list blocks: Defaults to an empty list (:py:data:`[]`). See "Example contents of :py:attr:`blocks`" below
                        for sample contents. Use the "power plant builder" method
                        :py:meth:`~plantpredict.powerplant.PowerPlant.add_block` to easily add a new block to the
                        attribute :py:attr:`blocks`. Subsequently use the methods
                        :py:meth:`~plantpredict.powerplant.PowerPlant.add_array`,
                        :py:meth:`~plantpredict.powerplant.PowerPlant.add_inverter`, and
                        :py:meth:`~plantpredict.powerplant.PowerPlant.add_dc_field` to build out the full power plant
                        hierarchical structure.

    Below are some samples of the more complex attributes that would be populated after calling
    :py:meth:`~plantpredict.powerplant.PowerPlant.get` on an existing power plant in PlantPredict. This also is a sample
    of what the contents might look like before creating a new powerplant with
    :py:meth:`~plantpredict.powerplant.PowerPlant.create` (or update an existing one with
    :py:meth:`~plantpredict.powerplant.PowerPlant.update`:

    .. container:: toggle

        .. container:: header

            Example contents of :py:attr:`transformers`

        .. container:: transformers

            .. code-block:: python

                powerplant.transformers = [{
                    "id": 23982,
                    "rating": 0.6,                # units [MVA]
                    "high_side_voltage": 4.0,     # units [kV]
                    "no_load_loss": 0.5,          # units [%]
                    "full_load_loss": 1.0,        # units [%]
                    "ordinal": 1
                }]

    .. container:: toggle

        .. container:: header

            Example contents of :py:attr:`transmission_lines`

        .. container:: transmission_lines

            .. code-block:: python

                powerplant.transmission_lines = [{
                    "id": 48373,
                    "length": 2.0,                             # units [km]
                    "resistance": 0.5,                         # units [Ohms/300 m]
                    "number_of_conductors_per_phase": 3,
                    "ordinal": 1
                }]

    .. container:: toggle

        .. container:: header

            Example contents of :py:attr:`blocks`

        .. container:: blocks

            .. code-block:: python

                from plantpredict.enumerations import TrackingTypeEnum, ModuleOrientationEnum, BacktrackingTypeEnum

                powerplant.blocks = [{
                    "name": 1,
                    "id": 57383,
                    "description": "Description of block."
                    "repeater": 5,
                    "energization_date": "2019-12-26T16:43:55.867Z",
                    "use_energization_date": True,
                    "arrays": [{
                        "name": 1,
                        "id": 22323,
                        "description": "Description of array.",
                        "repeater": 2,
                        "ac_collection_loss": 1.0,                              # units [%]
                        "das_load": 1.2,                                        # units [%]
                        "cooling_load": 0.8,                                    # units [%]
                        "additional_losses": 0.1,                               # units [%]
                        "match_total_inverter_kva": True,
                        "transformer_enabled": True,
                        "transformer_kva_rating": 600.0,                        # units [kVA]
                        "transformer_high_side_voltage": 34.7,                  # units [V]
                        "transformer_no_load_loss": 0.2,                        # units [%]
                        "transformer_full_load_loss": 0.7,                      # units [%]
                        "tracker_motor_losses": 0.1,                            # units [%]
                        "inverters": [{
                            "name": "A",
                            "id": 234290,
                            "description": "Description of inverter."
                            "repeater": 1,
                            "inverter_id": 242,
                            "inverter": {} # Inverter model contents
                            "setpoint_kw": 600.0,                                       # units [kW]
                            "power_factor": 1.0,
                            "kva_rating": 600.0,                                        # units [kW]
                            "dc_fields": [{
                                "name": 1,
                                "id": 235324,
                                "description": "Description of DC field.",
                                "repeater": 3,
                                "module_id": 749,
                                "module": {}  # Module model contents
                                "tracking_type": TrackingTypeEnum.FIXED_TILT,
                                "module_orientation": ModuleOrientationEnum.PORTRAIT,
                                "tables_removed_for_pcs": 0,
                                "modules_high": 4,
                                "modules_wide": 18,
                                "lateral_intermodule_gap": 0.02,                                    # units [m]
                                "vertical_intermodule_gap": 0.02,                                   # units [m]
                                "field_length": 20.0,                                               # units [m]
                                "field_width": 11.0,                                                # units [m]
                                "collector_bandwidth": 2.2,                                         # units [m]
                                "table_length": 6.7,                                                # units [m]
                                "tables_per_row": 3,
                                "post_to_post_spacing": 1.8,                                        # units [m]
                                "number_of_rows": 16,
                                "table_to_table_spacing": 0.05,                                     # units [m]
                                "module_azimuth": 180,                                              # units [degrees]
                                "module_tilt": 30,                                                  # units [degrees]
                                "tracking_backtracking_type": BacktrackingTypeEnum.TRUE_TRACKING,
                                "tracker_pitch_angle_d": 0,                                         # units [degrees]
                                "minimum_tracking_limit_angle_d": -60.0,                            # units [degrees]
                                "maximum_tracking_limit_angle_d": 60.0,                             # units [degrees]
                                "tracker_stow_angle": 0,                                            # units [degrees]
                                "post_height": 1.5,                                                 # units [m]
                                "structure_shading": 2.0,                                           # units [%]
                                "backside_mismatch": 1.0,                                           # units [%]
                                "field_dc_power": 800.0,                                            # units [kW]
                                "modules_wired_in_series": 10,
                                "number_of_series_strings_wired_in_parallel": 400,
                                "planned_module_rating": 325.0,                                     # units [W]
                                "sandia_conductive_coef": -3.47,
                                "sandia_convective_coef": -0.0594,
                                "cell_to_module_temp_diff": 3.0,                                    # units [deg-C]
                                "heat_balance_conductive_coef": 30.7,
                                "heat_balance_convective_coef": 0.0,
                                "module_mismatch_coefficient": 1.0,                                 # units [%]
                                "module_quality": 1.0,                                              # units [%]
                                "light_induced_degradation": 1.0,                                   # units [%]
                                "tracker_load_loss": 0.0,                                           # units [%]
                                "dc_wiring_loss_at_stc": 1.5,                                       # units [%]
                                "dc_health": 1.0,                                                   # units [%]
                            }],
                        }],
                    }],
                }]
    |
    """
    def create(self):
        """
        **POST** */Project/* :py:attr:`project_id` */Prediction/* :py:attr:`prediction_id` */PowerPlant*

        Creates a new power plant in the PlantPredict database with the attributes assigned to the instance of
        :py:class:`~plantpredict.powerplant.PowerPlant`. Automatically attaches it to a project/prediction existing in
        PlantPredict associated with the assigned values for :py:attr:`project_id` and
        :py:attr:`prediction_id`. Also automatically calculates the average power factor (plant design derate)
        based on the power factors of each inverter. See :py:class:`~plantpredict.powerplant.PowerPlant` documentation
        attributes required to successfully call this method.

        :return: Dictionary with contents :py:data:`{'is_successful': True}`.
        :rtype: dict
        """
        self._calculate_and_set_average_power_factor()

        self.create_url_suffix = "/Project/{}/Prediction/{}/PowerPlant".format(self.project_id, self.prediction_id)
        return super(PowerPlant, self).create()

    def get(self):
        """
        **GET** */Project/* :py:attr:`project_id` */Prediction/* :py:attr:`prediction_id` */PowerPlant*

        Retrieves an existing :py:class:`~plantpredict.powerplant.PowerPlant` from the PlantPredict database
        according to the values assigned for :py:attr:`project_id` and :py:attr:`prediction_id`, and
        automatically assigns all of its attributes to the object instance.

        :return: A dictionary containing all of the retrieved :py:class:`~plantpredict.powerplant.PowerPlant`
                 attributes. (Matches the contents of the attributes :py:attr:`__dict__` after calling this method).
        :rtype: dict
        """
        self.get_url_suffix = "/Project/{}/Prediction/{}/PowerPlant".format(self.project_id, self.prediction_id)
        return super(PowerPlant, self).get()

    def get_json(self):
        url_suffix = "/Project/{}/Prediction/{}/PowerPlant".format(self.project_id, self.prediction_id)
        create_request = requests.get(
            url=self.api.base_url + url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
           )
        return json.loads(create_request.content)

    def update_from_json(self, json_power_plant=None):
        url_suffix = "/Project/{}/Prediction/{}/PowerPlant".format(self.project_id, self.prediction_id)
        create_request = requests.put(
            url=self.api.base_url + url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=json_power_plant,
           )
        return json.loads(create_request.content)
    def calculate_dcfields(self, json_power_plant=None):
        url_suffix = "/Project/{}/Prediction/{}/CalculatePowerPlantFields".format(self.project_id, self.prediction_id)
        create_request = requests.post(
            url=self.api.base_url + url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=json_power_plant,
            )
        response = json.loads(create_request.content)
        return response
    def update_module(self, json_power_plant=None):
        url_suffix = "/Project/{}/Prediction/{}/CalculatePowerPlantFields".format(self.project_id, self.prediction_id)
        create_request = requests.post(
            url=self.api.base_url + url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=json_power_plant,
            )
        response = json.loads(create_request.content)  
        return self.update_from_json(response)
    def update(self):
        """
        **PUT** */Project/* :py:attr:`project_id` */Prediction/* :py:attr:`prediction_id` */PowerPlant*

        Updates an existing :py:class:`~plantpredict.powerplant.PowerPlant` entity in PlantPredict using the full
        attributes of the object instance. Calling this method is most commonly preceded by instantiating an
        :py:class:`~plantpredict.powerplant.PowerPlant` object with a particular :py:attr:`project_id` and
        :py:attr:`prediction_id` and calling :py:meth:`~plantpredict.powerplant.PowerPlant.get`, and changing any
        attributes locally.

        :return: Dictionary with contents :py:data:`{'is_successful': True}`.
        :rtype: dict
        """
        self.update_url_suffix = "/Project/{}/Prediction/{}/PowerPlant".format(self.project_id, self.prediction_id)
        return super(PowerPlant, self).update()

    def _calculate_sum_power_factors(self):
        """
        Calculates the sum of all the inverter power factors (design derate) in the power plant by iterating through
        each array of each block of :py:attr:`blocks`.

        :return: Sum of all power factors of each inverter in the power plant.
        :rtype: float
        """
        power_factors = []
        for block in self.blocks:
            for array in block['arrays']:
                for inverter in array['inverters']:
                    power_factors.append(inverter['power_factor'] * array['repeater'] * inverter['repeater'])

        return sum(power_factors)

    def _calculate_num_inverters(self):
        """
        Calculates the total number of inverters in the power plant by iterating through each array of each block of
        the attribute :py:attr:`blocks`.

        :return: Total number of inverters in the power plant.
        :rtype: int
        """
        num_inverters = []
        for block in self.blocks:
            for array in block['arrays']:
                for inverter in array['inverters']:
                    num_inverters.append(inverter['repeater'] * array['repeater'])

        return sum(num_inverters)

    def _calculate_and_set_average_power_factor(self):
        """
        Calculates the average power factor (design derate) of the power plant and sets it as the attribute
        :py:attr:`power_factor`.
        """

        total_power_factors = self._calculate_sum_power_factors()
        total_inverters = self._calculate_num_inverters()

        self.power_factor = 0.0 if total_inverters == 0 else total_power_factors / total_inverters

    def add_transformer(self, rating, high_side_voltage, no_load_loss, full_load_loss, ordinal):
        """
        Appends a transformer to the attribute :py:attr:`transformers` to model the system-level of the power plant.

        :param float rating: Transformer rating. Must be between :py:data:`0.1` and :py:data:`10000.0` - units
                             :py:data:`[MVA]`.
        :param float high_side_voltage: Transformer voltage. Must be between :py:data:`1.0` and :py:data:`1000.0` -
                                        units :py:data:`[kV]`.
        :param float no_load_loss: Transformer loss at no load. Must be between :py:data:`0.0` and :py:data:`10.0` -
                                   units :py:data:`[%]`.
        :param float full_load_loss: Transformer loss at full load. Must be between :py:data:`0.0` and :py:data:`10.0` -
                                     units :py:data:`[%]`.
        :param int ordinal: Order in sequence of :py:attr:`transformers` and :py:attr:`transmission_lines` where
                            :py:data:`1` represents the closest entity to the power plant/farthest entity from the
                            energy meter (1-indexed).
        """
        transformer = {
            "rating": rating,
            "high_side_voltage": high_side_voltage,
            "no_load_loss": no_load_loss,
            "full_load_loss": full_load_loss,
            "ordinal": ordinal
        }

        try:
            self.transformers.append(transformer)
        except AttributeError:
            self.transformers = [transformer]

    def add_transmission_line(self, length, resistance, number_of_conductors_per_phase, ordinal):
        """
        Appends a transmission line to the attribute :py:attr:`transmission_lines` to model the system-level of the
        power plant.

        :param float length: Length of transmission line. Must be between :py:data:`0.1` and :py:data:`100.0` - units
                     :py:data:`[km]`.
        :param float resistance: Transmission line resistivity (per 300m). Must be between :py:data:`0.001` and
                                 :py:data:`2` - units :py:data:`[Ohms/300m]`.
        :param int number_of_conductors_per_phase: Number of conductors per phase. Must be between :py:data:`1` and
                                                   :py:data:`10`.
        :param ordinal: Order in sequence of :py:attr:`transformers` and :py:attr:`transmission_lines` where
                        :py:data:`1` represents the closest entity to the power plant/farthest entity from the
                        energy meter (1-indexed).
        """
        transmission_line = {
            "length": length,
            "resistance": resistance,
            "number_of_conductors_per_phase": number_of_conductors_per_phase,
            "ordinal": ordinal
            }
        # append new transmission line, or if list doesn't yet exist, create it
        try:
            self.transmission_lines.append(transmission_line)
        except AttributeError:
            self.transmission_lines = [transmission_line]

    def _validate_block_name(self, block_name):
        """
        Checks that a given block with name `block_name` exists the power plant structure.

        :param int block_name: Name of block. Can be found as key `name` in each dictionary item of list `self.blocks`.
        :raises ValueError: Raised if no blocks in `self.blocks` have the name `block_name`.
        """
        if block_name not in [b['name'] for b in self.blocks]:
            raise ValueError("{} is not a valid block name in the existing power plant structure.".format(block_name))

    def _validate_array_name(self, block_name, array_name):
        """
        Checks that a given block with name `block_name` exists the power plant structure, and if so, that a given array
        with name `array_name` is a valid array in the block.

        :param int block_name: Name of block. Can be found as key `name` in each dictionary item of list `self.blocks`.
        :param int array_name: Name of array. Can be found as key `name` in each dictionary item of list
                               `self.blocks[i]["arrays"]`, where `i` is some valid integer index.
        :raises ValueError: Raised if no blocks in `self.blocks` have the name `block_name`. Also raised if `block_name`
                            is valid but there is no array in the block with name `array_name`.
        """
        self._validate_block_name(block_name)

        if array_name not in [a['name'] for a in self.blocks[block_name - 1]['arrays']]:
            raise ValueError("{} is not a valid array name in block {}.".format(array_name, block_name))

    def _validate_inverter_name(self, block_name, array_name, inverter_name):
        """
        Checks that a given block with name `block_name` exists the power plant structure, and if so, that a given array
        with name `array_name` is a valid array in the block, and if so that a given inverter with name `inverter_name`
        is a valid inverter in the array.

        :param int block_name: Name of block. Can be found as key `name` in each dictionary item of list `self.blocks`.
        :param int array_name: Name of array. Can be found as key `name` in each dictionary item of list
                               `self.blocks[i]["arrays"]`, where `i` is a valid integer index.
        :param str inverter_name: Name of inverter. Can be found as key `name` in each dictionary item of list
                                  `self.blocks[i]["arrays"][j]["inverters"]` where `i` and `j` are valid integer
                                  indices.
        :raises ValueError: Raised if no blocks in `self.blocks` have the name `block_name`. Also raised if `block_name`
                            is valid but there is no array in the block with name `array_name`.
        """
        self._validate_block_name(block_name)
        self._validate_array_name(block_name, array_name)

        if inverter_name not in [i['name'] for i in self.blocks[block_name - 1]['arrays'][array_name - 1]['inverters']]:
            raise ValueError(
                "'{}' is not a valid inverter name in array {} of block {}.".format(inverter_name, array_name,
                                                                                    block_name))

    @handle_refused_connection
    @handle_error_response
    def add_block(self, use_energization_date=False, energization_date=""):
        """
        A "power plant builder" helper method that creates a new block and appends it to the attribute
        :py:attr:`blocks`. Block naming is sequential (numerically) - for instance, if there are 2 existing blocks with
        names :py:data:`1` and :py:data:`2` (accessible via key :py:data:`name` on each block in list), the next block
        created by :py:meth:`~plantpredict.PowerPlant.powerplant.add_block` will automatically have :py:data:`name`
        equal to :py:data:`3`. This method does not currently account for the situation in which an existing power plant
        has blocks named non-sequentially.

        Note that this addition is not persisted to PlantPredict unless
        :py:meth:`~plantpredict.powerplant.PowerPlant.update` is subsequently called.

        :param bool use_energization_date: Enables use of energization date in power plant block. Defaults to
                                           :py:data:`False`.
        :param str energization_date: Timestamp representing energization date of block. Uses format
                                      :py:data:`2019-12-26T16:43:55.867Z` and defaults to :py:data:`""`.
        :return: Name of newly added block.
        :rtype: int
        """
        block = {
            "name": 1 if not self.blocks else len(self.blocks) + 1,
            "use_energization_date": use_energization_date,
            "energization_date": energization_date,
            "arrays": []
        }

        # if blocks list does not exit, create new list instead of appending
        try:
            self.blocks.append(block)
        except AttributeError:
            self.blocks = [block]

        return self.blocks[-1]["name"]

    @handle_refused_connection
    @handle_error_response
    def clone_block(self, block_id_to_clone):
        """
        A "power plant builder" helper method that clones (copies) an existing block (and all of its children
        arrays/inverters/DC fields) and appends it to attribute :py:attr:`blocks`. Particularly useful when you want to
        create a new block that is similar to an existing block. Block naming is sequential (numerically) - for
        instance, if there are 2 existing blocks with names :py:data`1` and :py:data:`2` (accessible via key
        :py:data:`name` on each block in list), the next block created by
        :py:meth:`~plantpredict.powerplant.PowerPlant.clone_block` will automatically have :py:data`name` equal to
        :py:data`3`. This method does not currently account for the situation in which an existing power plant has
        blocks named non-sequentially.

        Note that this addition is not persisted to PlantPredict unless
        :py:meth:`~plantpredict.powerplant.PowerPlant.update` is subsequently called.

        :param int block_id_to_clone: Unique identifier of the block you wis you clone. Can be found in the relevant
                                      block dictionary (in list :py:attr:`self.blocks`) with key :py:data:`id`.
        :return: Name of newly cloned block.
        :rtype: int
        """
        block_to_clone = [b for b in self.blocks if b['id'] == block_id_to_clone][0]
        block_copy = copy.deepcopy(block_to_clone)
        block_copy["name"] = len(self.blocks) + 1

        block_copy.pop("id")  # remove id from copied block so that PlantPredict will assign a new unique id when block is created
        if (block_copy["nodal_export_options"] is not None):
            block_copy["nodal_export_options"].pop("id")  # remove id from nodal export options of copied block so that PlantPredict will assign a new unique id when block is created
        
        self.blocks.append(block_copy)
        self.update()

        return self.blocks[-1]["name"]

    @handle_refused_connection
    @handle_error_response
    def add_array(self, block_name, transformer_enabled=True, match_total_inverter_kva=True,
                  transformer_kva_rating=None, repeater=1, ac_collection_loss=1, das_load=800, cooling_load=0.0,
                  additional_losses=0.0, transformer_high_side_voltage=34.5, transformer_no_load_loss=0.2,
                  transformer_full_load_loss=0.7, description=""):
        """
        A "power plant builder" helper method that adds an array to the block specified by :py:data:`block_name` on the
        :py:class:`~plantpredict.powerplant.PowerPlant`. Array naming is sequential (numerically) - for instance, if
        there are 2 existing arrays with names :py:data:`1` and :py:data:`2` (accessible via key :py:data:`name` for a
        given array dictionary), the next array created by :py:meth:`~plantpredict.powerplant.PowerPlant.add_array` will
        automatically have :py:data:`name` equal to :py:data:`3`. This method does not currently account for the
        situation in which an existing power plant has arrays named non-sequentially.

        Note that this addition is not persisted to PlantPredict unless
        :py:meth:`~plantpredict.powerplant.PowerPlant.update` is subsequently called.

        :param int block_name: Name (1-indexed integer) of the parent block to add the array to. Can be found in the
                               relevant block dictionary (in attribute :py:attr:`blocks`) with key :py:data:`id`. This
                               value is returned for a new block when you create one with
                               :py:meth:`~plantpredict.powerplant.PowerPlant.add_block`. Must be between :py:data:`1`
                               and :py:data:`99`.
        :param bool transformer_enabled: If :py:data:`True`, enables a medium-voltage (MV) transformer for the array.
                                         Defaults to :py:data:`True`.
        :param bool match_total_inverter_kva: If :py:data:`True`, the transformer size will match the total inverter kVA
                                              of the inverter behind the transformer, and the input
                                              :py:data:`transformer_kva_rating` won't be used. Defaults to
                                              :py:data:`True`.
        :param transformer_kva_rating: User-specified transformer kVA rating. Only used if
                                       :py:data:`match_total_inverter_kva` is set to :py:data:`False`. Defaults to
                                       :py:data:`None`. Must be between :py:data:`0` and :py:data:`20000` - units
                                       :py:data:`[kVA]`.
        :type transformer_kva_rating: float, None
        :param int repeater: Number of identical arrays of this type in the parent block.  Defaults to :py:data:`1`.
                             Must be between :py:data:`1` and :py:data:`10000`.
        :param float ac_collection_loss: Accounts for ohmic losses in the AC wiring between the array and parent block.
                                         Defaults to :py:data:`1`.Must be between :py:data:`0` and :py:data:`30` - units
                                         :py:data:`[%]`.
        :param float das_load: Accounts for parasitic losses due to the data acquisition system (DAS). Can also be used
                               for general time-constant parasitic loss accounting. Defaults to :py:data:`800`. Must be
                               between :py:data:`0` and :py:data:`5000` - units :py:data:`[W]`.
        :param float cooling_load: Accounts for losses from the power conditioning system (PCS) shelter cooling system.
                                   Defaults to :py:data:`0.0`. Must be between :py:data:`0` and :py:data:`5000` - units
                                   :py:data:`[W]`.
        :param float additional_losses: Additional night time losses. Defaults to :py:data:`0`. Must be between
                                        :py:data:`0` and :py:data:`20000` - units :py:data:`[W]`.
        :param float transformer_high_side_voltage: Transformer high side voltage (the AC collection line voltage
                                                    defines the high-side of a MV inverter). Defaults to
                                                    :py:data:`34.5`. Must be between :py:data:`0` and :py:data:`66` -
                                                    units :py:data:`[V]`.
        :param float transformer_no_load_loss: Accounts for transformer losses with no load. Defaults to :py:data:`0.2`.
                                               Must be between :py:data:`0` and :py:data:`10` - units :py:data:`[%]`.
        :param float transformer_full_load_loss: Accounts for transformer losses with full load. Defaults to
                                                 :py:data:`0.7`. Must be between :py:data:`0` and :py:data:`10` - units
                                                 :py:data:`[%]`.
        :param str description: Description of the array. Must be :py:data:`250` characters or less. Defaults to
                                :py:data:`""`.
        :raises ValueError: Raised if :py:data:`block_name` is not a valid block name in the existing power plant.
        :return: The name of the newly added array.
        :rtype: int
        """
        self._validate_block_name(block_name)

        array = {
            "name": len(self.blocks[block_name - 1]["arrays"]) + 1,
            "repeater": repeater,
            "ac_collection_loss": ac_collection_loss,
            "das_load": das_load,
            "cooling_load": cooling_load,
            "additional_losses": additional_losses,
            "transformer_enabled": transformer_enabled,
            "match_total_inverter_kva": match_total_inverter_kva,
            "transformer_high_side_voltage": transformer_high_side_voltage,
            "transformer_no_load_loss": transformer_no_load_loss,
            "transformer_full_load_loss": transformer_full_load_loss,
            "inverters": [],
            "description": description
        }
        if not match_total_inverter_kva:
            array.update({"transformer_kva_rating": transformer_kva_rating})

        self.blocks[block_name - 1]["arrays"].append(array)

        return self.blocks[block_name - 1]["arrays"][-1]["name"]

    @handle_refused_connection
    @handle_error_response
    def _get_inverter_apparent_power(self, inverter_id):
        """
        Returns the apparent power of an inverter specified by its unique identifier.

        :param int inverter_id: Unique identifier of an Inverter in the PlantPredict Inverter database.
        :return: Apparent power of inverter model - units `[kVA]`.
        :rtype: float
        """
        inverter = self.api.inverter(id=inverter_id)
        inverter.get()

        return inverter.apparent_power

    @handle_refused_connection
    @handle_error_response
    def _get_inverter_kva_rating(self, inverter_id):
        """
        Gets the inverters kVA rating based on the elevation and 99.6 Cooling Temperature (which comes
        from the ASHRAE station nearest to the latitude and longitude) of the :py:class:`~plantpredict.project.Project`
        corresponding to `self.project_id`.

        :param int inverter_id: Unique identifier of an Inverter in the PlantPredict Inverter database.
        :return: Kilovolt-Ampere rating, used to rate/size the transformer of a power plant - units py:data:`[kVA]`.
        :rtype: float
        """
        # retrieve ASHRAE station based on latitude and longitude of project associated with power plant
        project = self.api.project(id=self.project_id)
        project.get()
        prediction = self.api.prediction(id=self.prediction_id, project_id=self.project_id)
        prediction.get()
        ashrae = self.api.ashrae(
            latitude=project.latitude,
            longitude=project.longitude,
            station_name=prediction.ashrae_station
        )
        ashrae.get_station()

        # use the kVA endpoint to calculate the kVA with elevation and 99.6 cooling temp of nearest ASHRAE station
        inverter = self.api.inverter(id=inverter_id)
        response = inverter.get_kva(
            elevation=project.elevation,
            temperature=ashrae.cool_996,
            use_cooling_temp=self.use_cooling_temp
        )
        return response['kva']

    @staticmethod
    def _validate_inverter_setpoint_inputs(setpoint_kw, power_factor, kva_rating):
        """
        Ensures valid inputs for :py:data:`setpoint_kw` and :py:data:`power_factor`. In general, ensures that the ratio
        `power_factor = setpoint_kw / kva_rating` is maintained, while kva_rating is held constant.

        :param float, None setpoint_kw: Inverter setpoint. Must be between :py:data:`1` and :py:data:`10000` - units
                                        `[kW]`.
        :param float power_factor: The ratio of the power that can be used and the product of the operating current and
                                   voltage. Must be between :py:data:`0` and :py:data:`1`, where `1` is a "unity" power
                                   factor.
        :param float kva_rating: Inverter kVA rating.
        :raises ValueError: Raised if :py:data:`setpoint_kw` is not `None` and :py:data:`power_factor` is not `1.0`.
        :return: Valid inverter setpoint and power factor (design derate).
        :rtype: tuple
        """
        # if setpoint isn't provided, calculate by multiplying design derate by kvarating
        if setpoint_kw is None:
            setpoint_kw = power_factor * kva_rating

        # if setpoint is provided, recalculate design derate as the ratio of setpoint ot kva rating
        elif (setpoint_kw is not None) and (power_factor == 1.0):
            power_factor = setpoint_kw / kva_rating

        # setpoint cannot be provided when a non-unity power factor is provided (since kva rating is constant)
        elif (setpoint_kw is not None) and (power_factor != 1.0):
            raise ValueError("setpoint_kw can not be specified while a non-unity (non-1.0) power factor is specified.")

        return setpoint_kw, power_factor

    @handle_refused_connection
    @handle_error_response
    def add_inverter(self, block_name, array_name, inverter_id, setpoint_kw=None, power_factor=1.0, repeater=1):
        """
        A "power plant builder" helper method that adds an inverter to an array specified by :py:data:`array_name`,
        which is a child of a block specified by :py:data:`block_name` on the
        :py:class:`~plantpredict.powerplant.PowerPlant`. Inverter naming is sequential (alphabetically) - for instance,
        if there are 2 existing inverters with names :py:data:`"A"` and :py:data:`"B"` (accessible via key
        :py:data:`name` for a given inverter dictionary), the next array created by
        :py:meth:`~plantpredict.powerplant.PowerPlant.add_inverter` will automatically have :py:data:`name` equal to
        :py:data:`"C"`. This method does not currently account for the situation in which an existing power plant has
        inverters named non-sequentially.

        The inverter :py:data:'kva_rating` will be set based on the power plant-level attribute
        :py:attr:`use_cooling_temp`. If :py:attr:`use_cooling_temp` is :py:data:`True`, this value is automatically
        calculated based on the 99.6 cooling temperature of the nearest ASHRAE station to the corresponding
        :py:class:`~plantpredict.project.Project` (as specified by the attribute :py:attr:`project_id`), the elevation
        of the :py:class:`~plantpredict.project.Project`, and the elevation/temperature curves of the inverter model
        specified by :py:data:`inverter_id`. If :py:attr:`use_cooling_temp` is :py:data:`False`, then
        :py:data:`kva_rating` is set as the :py:attr:`apparent_power` of the inverter model specified by
        :py:data:`inverter_id`.

        Note that this addition is not persisted to PlantPredict unless
        :py:meth:`~plantpredict.powerplant.PowerPlant.update` is subsequently called.

        :param int block_name: Name (1-indexed integer) of the parent block to add the inverter to. Can be found in the
                               relevant block dictionary (in attribute :py:attr:`blocks`) with key :py:data:`id`.  This
                               value is returned for a new block when you create one with
                               :py:meth:`~plantpredict.powerplant.PowerPlant.add_block`. Must be between :py:data:`1`
                               and :py:data:`99`.
        :param int array_name: Name (1-indexed integer) of the parent array to add the inverter to. This value is
                               returned for a new array when you create one with
                               :py:meth:`~plantpredict.powerplant.PowerPlant.add_array`. Must be between :py:data:`1`
                               and :py:data:`99`.
        :param int inverter_id: Unique identifier of an inverter model in the PlantPredict Inverter database to use.
        :param setpoint_kw: Inverter setpoint. Must be between :py:data:`1` and :py:data:`10000` - units
                            :py:data:`[kW]`. If left as default (:py:data:`None`), will be automatically calculated as
                            the product between :py:data:`power_factor` and the inverter kVA rating.
        :type setpoint_kw: float, None
        :param float power_factor: The ratio of the power that can be used and the product of the operating current and
                                   voltage (also referred to as design derate). Must be between :py:data:`0` and
                                   :py:data:`1`, where :py:data:`1` is a "unity" power factor. Defaults to
                                   :py:data:`1.0`.
        :param int repeater: Number of identical inverters of this type in the parent array. Must be between
                             :py:data:`1` and :py:data:`10000`. Defaults to :py:data:`1`.
        :raises ValueError: Raised if :py:data:`block_name` is not a valid block name in the existing power plant, or if
                            the :py:data:`block_name` is valid but :py:data:`array_name` is not a valid array name in
                            the block. Also raised if :py:data:`setpoint_kw` is not :py:data:`None` and
                            :py:data:`power_factor` is not :py:data:`1.0`.
        :return: The name of the newly added inverter.
        :rtype: str
        """
        # validate and prepare inverter parameters
        self._validate_array_name(block_name, array_name)

        kva_rating = (self._get_inverter_kva_rating(inverter_id) if self.use_cooling_temp
                      else self._get_inverter_apparent_power(inverter_id))
        setpoint_kw, power_factor = self._validate_inverter_setpoint_inputs(setpoint_kw, power_factor, kva_rating)

        self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"].append({
            "name": chr(ord("A") + len(self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"])),
            "repeater": repeater,
            "inverter_id": inverter_id,
            "setpoint_kw": setpoint_kw,
            "power_factor": power_factor,
            "dc_fields": [],
            "kva_rating": kva_rating
        })

        return self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][-1]["name"]

    def _get_default_module_azimuth_from_latitude(self):
        """
        Determines the default module azimuth (the orientation of the entire DC field) based on the latitude of the
        :py:class:`~plantpredict.project.Project` associated with the :py:class:`~plantpredict.powerplant.PowerPlant`
        (using :py:attr:`self.project_id`). By default, the DC field is set to be oriented south if above equator. The
        convention is 0.0 degrees for North-facing arrays.

        :return: Default azimuth, :py:data:`180.0` if latitude is above equator, otherwise :py:data:`0.0` - units
                 `[degrees]`.
        :rtype: float
        """
        p = self.api.project(id=self.project_id)
        p.get()
        azimuth = 180.0 if p.latitude >= 0.0 else 0.0

        return azimuth

    @staticmethod
    def _calculate_collector_bandwidth(module_width, module_length, module_orientation, modules_high,
                                       vertical_intermodule_gap):
        """
        Calculates the total width/depth of each table/row of modules in the DC field. The collector bandwidth is
        calculated by multiplying the number of modules high (number of ranks) by each module's vertical dimension,
        which is dependent on if the modules is oriented in portrain or landscape, and then adding the vertical space
        between each module.

        :param float module_width: Width of each individual module in DC field. Must be between :py:data:`0` and
                                   :py:data:`10000` - units `[mm]`.
        :param float module_length: Length of each individual module in DC field. Must be between :py:data:`0` and
                                    :py:data:`10000` - units `[mm]`.
        :param int module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field.
                                       Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :param int modules_high: Number of modules high per table (number of ranks). Must be between :py:data:`1` and
                                 :py:data:`50`.
        :param vertical_intermodule_gap: vertical gap between each module on the mounting structure. Must be between
                                         :py:data:`0` and py:data:`1` - units `[m]'.
        :return: Collector bandwidth for a table in the DC field - units `[m]`.
        :rtype: float
        """
        module_bandwidth = module_width if module_orientation == ModuleOrientationEnum.LANDSCAPE else module_length

        return modules_high * module_bandwidth / 1000.0 + (modules_high - 1) * vertical_intermodule_gap

    @staticmethod
    def _calculate_table_length(modules_wide, module_orientation, module_length, module_width, lateral_intermodule_gap):
        """
        Calculates the length of each table (mounting structure) in meters, for a particular DC field.

        :param int modules_wide: Number of modules across per table. Must be between :py:data:`1` and :py:data:`100`.
        :param int module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field.
                                       Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :param float module_length: Length of each individual module in DC field. Must be between :py:data:`0` and
                                    :py:data:`10000` - units `[mm]`.
        :param float module_width: Width of each individual module in DC field. Must be between :py:data:`0` and
                                   :py:data:`10000` - units `[mm]`.
        :param float lateral_intermodule_gap: Lateral gap between each module on the mounting structure. Must be between
                                              :py:data:`0` and py:data:`1` - units `[m]'.
        :return: Length of each table (mounting structure) for DC field - units `[m]`.
        :rtype: float
        """
        # choose the relevant module dimension (length or width) based on the module orientation
        module_dimension = module_length / 1000.0 if module_orientation == ModuleOrientationEnum.LANDSCAPE \
            else module_width / 1000.0

        return modules_wide*module_dimension + lateral_intermodule_gap*(modules_wide - 1)

    @staticmethod
    def _calculate_tables_per_row(field_dc_power, planned_module_rating, modules_high, modules_wide,
                                  number_of_rows, tables_removed_for_pcs=0):
        """
        Calculates the number of tables (mounting structures) across per row of a DC field. Rounds the result up to
        the nearest next integer (ceiling function), as would be done automatically in the backend of PlantPredict.

        :param float field_dc_power: DC capacity of the DC field. Must be between :py:data:`1` and :py:data:`20000` -
                                     units `[kW]`.
        :param float planned_module_rating: Nameplate rating of each individual module in the DC field. Must be between
                                            :py:data:`10` and :py:data:`1000` - units `[W]`.
        :param int modules_high: Number of modules high per table (number of ranks). Must be between :py:data:`1` and
                                 :py:data:`50`.
        :param int modules_wide: Number of modules across per table. Must be between :py:data:`1` and :py:data:`100`.
        :param int number_of_rows: Number of rows of tables in DC field. Must be between :py:data:`1` and
                                   :py:data:`10000`.
        :param float tables_removed_for_pcs: Number of tables removed in DC field to make room for its power
                                             conditioning system (PCS). Must be between :py:data:`0` and :py:data:`50`.
                                             Defaults to :py:data:`0`.
        :return: Number of tables across per row of DC field.
        :rtype: int
        """
        module_count = 1000*field_dc_power / planned_module_rating
        modules_per_table = modules_high * modules_wide            # note: only a frontend value
        total_tables = module_count / modules_per_table            # note: only a frontend value
        tables_per_row = (total_tables + tables_removed_for_pcs) / number_of_rows

        return math.ceil(tables_per_row)    # rounds up to next greater integer

    @staticmethod
    def _calculate_dc_field_size_by_collector_bandwidth(number_of_rows, post_to_post_spacing, collector_bandwidth):
        """
        Calculates the DC field dimension from the front of the first table to the back of the last table (the first
        row to the last row).

        :param int number_of_rows: Number of rows of tables in DC field. Must be between :py:data:`1` and
                                     :py:data:`10000`.
        :param float post_to_post_spacing: Row spacing. Must be between :py:data:`0.0` and :py:data:`50.0` - units
                                           :py:data:`[m]`.
        :param float collector_bandwidth: The total width/depth of each table/row of modules in the DC field. Must be
                                          between :py:data:`0` and :py:data:`30` - units `[m]`.
        :return: Dimension of DC field in the "front to back" direction - units `[m]`.
        :rtype: float
        """
        return post_to_post_spacing*(number_of_rows - 1) + collector_bandwidth

    @staticmethod
    def _calculate_dc_field_size_by_tables_per_row(tables_per_row, module_orientation, module_length, module_width,
                                                   lateral_intermodule_gap, modules_wide):
        """
        Calculates the DC field dimension along each row of tables/modules.

        :param float tables_per_row: Number of tables wide per row in the DC field.
        :param int module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field.
                                       Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :param float module_length: Length of each individual module in DC field. Must be between :py:data:`0` and
                                    :py:data:`10000` - units `[mm]`.
        :param float module_width: Width of each individual module in DC field. Must be between :py:data:`0` and
                                   :py:data:`10000` - units `[mm]`.
        :param float lateral_intermodule_gap: Lateral gap between each module on the mounting structure. Must be between
                                              :py:data:`0` and py:data:`1` - units `[m]'.
        :param int modules_wide: Number of modules across per table. Must be between :py:data:`1` and :py:data:`100`.
        :return: Dimension of DC field in the "side to side" direction - units `[m]`.
        :rtype: float
        """
        module_size = module_length / 1000.0 if module_orientation == ModuleOrientationEnum.LANDSCAPE \
            else module_width / 1000.0

        return (modules_wide * tables_per_row * (module_size + lateral_intermodule_gap)) - lateral_intermodule_gap

    def _calculate_dc_field_length(self, tables_per_row, module_orientation, module_length, module_width,
                                   lateral_intermodule_gap, modules_wide, tracking_type, number_of_rows,
                                   post_to_post_spacing, collector_bandwidth):
        """
        Calculates the DC field length dimension, which is the "side to side" dimension across each row for a horizontal
        tracker array, and the "front to back" dimension from the front row to the back row of tables for a fixed tilt
        array.

        :param int tables_per_row: Number of tables wide per row in the DC field.
        :param int module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field.
                                       Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :param float module_length: Length of each individual module in DC field. Must be between :py:data:`0` and
                                    :py:data:`10000` - units `[mm]`.
        :param float module_width: Width of each individual module in DC field. Must be between :py:data:`0` and
                                   :py:data:`10000` - units `[mm]`.
        :param float lateral_intermodule_gap: Lateral gap between each module on the mounting structure. Must be between
                                              :py:data:`0` and py:data:`1` - units `[m]'.
        :param int modules_wide: Number of modules across per table. Must be between :py:data:`1` and :py:data:`100`.
        :param int tracking_type: Represents the tracking type/mounting structure (Fixed Tilt, Tracker, etc.) of the DC
                                  field. Use :py:class:`~plantpredict.enumerations.TrackingTypeEnum`.
        :param int number_of_rows: Number of rows of tables in DC field. Must be between :py:data:`1` and
                                     :py:data:`10000`.
        :param float post_to_post_spacing: Row spacing. Must be between :py:data:`0.0` and :py:data:`50.0` - units
                                           :py:data:`[m]`.
        :param float collector_bandwidth: The total width/depth of each table/row of modules in the DC field. Must be
                                          between :py:data:`0` and :py:data:`30` - units `[m]`.
        :return: DC field length - units py:data:`[m]`.
        :rtype: float
        """
        if tracking_type == TrackingTypeEnum.HORIZONTAL_TRACKER:
            return self._calculate_dc_field_size_by_tables_per_row(tables_per_row, module_orientation, module_length,
                                                                   module_width, lateral_intermodule_gap, modules_wide)

        return self._calculate_dc_field_size_by_collector_bandwidth(number_of_rows, post_to_post_spacing,
                                                                    collector_bandwidth)

    def _calculate_dc_field_width(self, tracking_type, number_of_rows, post_to_post_spacing, collector_bandwidth,
                                  tables_per_row, module_orientation, module_length, module_width,
                                  lateral_intermodule_gap, modules_wide):
        """
        Calculates the DC field width dimension, which is the "side to side" dimension across each row for a horizontal
        fixed tilt array, and the "front to back" dimension from the front row to the back row of tables for a tracker
        array.

        :param int tracking_type: Represents the tracking type/mounting structure (Fixed Tilt, Tracker, etc.) of the DC
                                  field. Use :py:class:`~plantpredict.enumerations.TrackingTypeEnum`.
        :param int number_of_rows: Number of rows of tables in DC field. Must be between :py:data:`1` and
                                     :py:data:`10000`.
        :param float post_to_post_spacing: Row spacing. Must be between :py:data:`0.0` and :py:data:`50.0` - units
                                           :py:data:`[m]`.
        :param float collector_bandwidth: The total width/depth of each table/row of modules in the DC field. Must be
                                          between :py:data:`0` and :py:data:`30` - units `[m]`.
        :param int tables_per_row: Number of tables wide per row in the DC field.
        :param int module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field.
                                       Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :param float module_length: Length of each individual module in DC field. Must be between :py:data:`0` and
                                    :py:data:`10000` - units `[mm]`.
        :param float module_width: Width of each individual module in DC field. Must be between :py:data:`0` and
                                   :py:data:`10000` - units `[mm]`.
        :param float lateral_intermodule_gap: Lateral gap between each module on the mounting structure. Must be between
                                              :py:data:`0` and py:data:`1` - units `[m]'.
        :param int modules_wide: Number of modules across per table. Must be between :py:data:`1` and
                                       :py:data:`100`.
        :return: DC field width - units :py:data:`[m]`.
        :rtype: float
        """
        if tracking_type == TrackingTypeEnum.HORIZONTAL_TRACKER:
            return self._calculate_dc_field_size_by_collector_bandwidth(number_of_rows, post_to_post_spacing,
                                                                        collector_bandwidth)

        return self._calculate_dc_field_size_by_tables_per_row(tables_per_row, module_orientation, module_length,
                                                               module_width, lateral_intermodule_gap, modules_wide)

    @staticmethod
    def _validate_dc_field_sizing(field_dc_power, number_of_series_strings_wired_in_parallel, planned_module_rating,
                                  modules_wired_in_series):
        """
        Calculates equivalent number of strings in parallel for a given DC capacity, or calculates an equivalent DC
        capacity for a given number of strings in parallel for the DC field. Returns both the provided and calculated
        values.

        :param float, None field_dc_power: DC capacity of the DC field. Must be `None` if
                                           `number_of_seres_strings_wired_in_parallel` is not `None`, otherwise must be
                                           between :py:data:`1` and :py:data:`20000` - units `[kW]`.
        :param float, None number_of_series_strings_wired_in_parallel: Number of strings of modules electrically
                                                                       connected in parallel in the DC field. Must be
                                                                       `None` if `field_dc_power` is not `None`,
                                                                       otherwise must be between :py:data:`1` and
                                                                       :py:data`10000`.
        :param float planned_module_rating: Nameplate rating of each individual module in the DC field. Must be between
                                            :py:data:`10` and :py:data:`1000` - units `[W]`.
        :param int modules_wired_in_series: The number of modules electrically connected in series in a string. Must be
                                            be between :py:data:`1` and :py:data:`100`.
        :raises ValueError: Raised if both `field_dc_power` and `number_of_series_strings_wired_in_parallel` are `None`
                            or are both not `None`.
        :return: Field DC Power (units `[kW]`) and equivalent number of series strings wired in parallel in DC field.
        :rtype: tuple
        """
        if (field_dc_power is not None) and (number_of_series_strings_wired_in_parallel is not None):
            raise ValueError("Both field_dc_power and number_of_series_strings_wired_in_parallel are not None. Only "
                             "one of these values can be specified (and the other will be calculated).")

        # calculates number of strings from field dc power if field dc power specified
        elif (field_dc_power is not None) and (number_of_series_strings_wired_in_parallel is None):
            number_of_series_strings_wired_in_parallel = 1000 * field_dc_power / \
                                                         (planned_module_rating * modules_wired_in_series)

        # calculates field dc power from number of strings if number of strings specified
        elif (number_of_series_strings_wired_in_parallel is not None) and (field_dc_power is None):
            field_dc_power = number_of_series_strings_wired_in_parallel * \
                             (planned_module_rating * modules_wired_in_series) / 1000.0

        else:
            raise ValueError("Both field_dc_power and number_of_series_strings_wired_in_parallel are None. One "
                             "of these variables must be specified, and the other will be calculated.")

        return field_dc_power, number_of_series_strings_wired_in_parallel

    @handle_refused_connection
    @handle_error_response
    def calculate_post_to_post_spacing_from_gcr(self, ground_coverage_ratio, module_id, modules_high,
                                                module_orientation=None, vertical_intermodule_gap=0.02):
        """
        Useful helper method for calculating :py:attr:`post_to_post_spacing` based on a desired ground coverage ratio
        (GCR). :py:attr:`post_to_post_spacing` is a required input for
        :py:meth:`~plantpredict.powerplant.PowerPlant.add_dc_field`.

        :param float ground_coverage_ratio: Ratio of collector bandwidth to row spacing - units :py:data:`[decimal]`.
        :param int module_id: Unique identifier of the module to be used in the DC field.
        :param int modules_high: Number of modules high per table (number of ranks). Must be between :py:data:`1` and
                                 :py:data:`50`.
        :param module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field. If
                                   left as default (:py:data:`None`), is automatically set as the
                                   :py:attr:`module_orientation` of the module model specified by :py:data:`module_id`.
                                   Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :type module_orientation: int, None
        :param float vertical_intermodule_gap: Vertical gap between each module on the mounting structure. Defaults to
                                               :py:data:`0.02`. Must be between :py:data:`0` and py:data:`1` - units
                                               :py:data:`[m]'.
        :return: Post to post spacing (row spacing) of DC field - units :py:data:`[m]`.
        :rtype: float
        """
        m = self.api.module(id=module_id)
        m.get()

        collector_bandwidth = self._calculate_collector_bandwidth(
            module_width=m.width,
            module_length=m.length,
            module_orientation=module_orientation if module_orientation is not None else m.default_orientation,
            modules_high=modules_high,
            vertical_intermodule_gap=vertical_intermodule_gap
        )

        return collector_bandwidth / ground_coverage_ratio

    @staticmethod
    def calculate_field_dc_power_from_dc_ac_ratio(dc_ac_ratio, inverter_setpoint):
        """
        Useful helper method for sizing the DC field capacity (:py:attr:`field_dc_power`) based on a desired DC AC ratio
        and known inverter setpoint. :py:attr:`field_dc_power` is a required input for
        :py:meth:`~plantpredict.powerplant.PowerPlant.add_dc_field`.

        :param float dc_ac_ratio: Ratio of DC capacity of DC field to the AC capacity/inverter setpoint.
        :param float inverter_setpoint: Setpoint of parent inverter to the DC field. Can be found with key
                                        :py:data:`setpoint_kw` in the dictionary representing the inverter. Must be
                                        between :py:data:`1` and :py:data:`10000` - units :py:data:`[kW]`.
        :return: DC capacity for a DC field - units :py:data:`[kW]`.
        :rtype: float
        """
        return dc_ac_ratio*inverter_setpoint

    @staticmethod
    def _validate_mounting_structure_parameters(tracking_type, module_tilt, tracking_backtracking_type):
        """
        Ensures that if the DC field uses a fixed tilt mounting structure, that a tilt angle is provided, and if it
        uses a horizontal tracker mounting structure, that a backtracking type is provided.

        :param int tracking_type: Represents the tracking type/mounting structure (Fixed Tilt, Tracker, etc.) of the DC
                                  field. Use :py:class:`~plantpredict.enumerations.TrackingTypeEnum`.
        :param module_tilt: Tilt angle of modules in DC Field for a fixed tilt array. Must be between :py:data:`0` and
                            :py:data:`90` - units `[degrees]`.
        :type module_tilt: float, None
        :param tracking_backtracking_type: Represents the backtracking algorithm (True-Tracking or Backtracking) used in
                                           DC Field. Use :py:class:`~plantpredict.enumerations.BacktrackingTypeEnum`.
        :type tracking_backtracking_type: int, None
        :raises ValueError: Raised if the `tracking_type` is
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.FIXED_TILT` and `module_tilt` is
                            `None`, or if `tracking_type` is
                            `~plantpredict.enumerations.TrackingTypeEnum.HORIZONTAL_TRACKER` and
                            `tracking_backtracking_type` is `None`.
]        """
        if (tracking_type == TrackingTypeEnum.FIXED_TILT) and (module_tilt is None):
            raise ValueError("The input module_tilt is required for a fixed tilt DC field.")
        elif (tracking_type == TrackingTypeEnum.HORIZONTAL_TRACKER) and (tracking_backtracking_type is None):
            raise ValueError("The input tracking_backtracking_type is required for a horizontal tracker DC field.")

    @staticmethod
    def _calculate_default_post_height(tracking_type, collector_bandwidth, module_tilt, minimum_tracking_limit_angle_d,
                                       maximum_tracking_limit_angle_d):
        """
        First calculates a default post height value. Then, the maximum of the calculated post height and 1.5 is
        returned.

        :param int tracking_type: Represents the tracking type/mounting structure (Fixed Tilt or Tracker) of the DC
                                  field. Use :py:class:`~plantpredict.enumerations.TrackingTypeEnum`. (Seasonal Tilt
                                  currently not supported in this package).
        :param float collector_bandwidth: The total width/depth of each table/row of modules in the DC field. Must be
                                          between :py:data:`0` and :py:data:`30` - units `[m]`.
        :param float, None module_tilt: Tilt angle of modules in DC Field for a fixed tilt array. Defaults to `None`.
                                        Non-null value required required if :py:data:`tracking_type` is equal to
                                        :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.FIXED_TILT`, and must be
                                        between :py:data:`0` and :py:data:`90` - units `[degrees]`.
        :param float minimum_tracking_limit_angle_d: Minimum tracking angle for horizontal tracker array. Defaults to
                                                     :py:data:`-60.0`. Must be between :py:data:`-90` and :py:data:`0` -
                                                     units `[degrees]`.
        :param float maximum_tracking_limit_angle_d: Maximum tracking angle for horizontal tracker array. Defaults to
                                                     :py:data:`60.0`. Must be between :py:data:`0` and :py:data:`90` -
                                                     units `[degrees]`.
        :return: Default post height value.
        :rtype: float
        """
        tilt = module_tilt if tracking_type == TrackingTypeEnum.FIXED_TILT else max(
            abs(minimum_tracking_limit_angle_d), abs(maximum_tracking_limit_angle_d)
        )
        post_height = ((collector_bandwidth * np.sin(np.deg2rad(tilt))) / 2) + 1

        return max(post_height, 1.5)

    @staticmethod
    def _calculate_modules_wide(strings_wide, modules_wired_in_series):
        """
        Calculates number of modules wide across table by multiplying the number of strings wide and modules per string.

        :param int strings_wide: Number of strings across per table. Multiplied by :py:data:`modules_wired_in_series` to
                                 determine :py:attr:`modules_wide`. Must result in :py:attr:`modules_wide` between
                                 :py:data:`1` and  :py:data:`100`. Defaults to :py:data:`1`.
        :param int modules_wired_in_series: The number of modules electrically connected in series in a string.
        :return: Modules wide across per table.
        :rtype: int
        """
        return strings_wide * modules_wired_in_series

    @handle_refused_connection
    @handle_error_response
    def add_dc_field(self, block_name, array_name, inverter_name, module_id, tracking_type, modules_high,
                     modules_wired_in_series, post_to_post_spacing, number_of_rows=1, strings_wide=1,
                     field_dc_power=None, number_of_series_strings_wired_in_parallel=None, module_tilt=None,
                     module_orientation=None, module_azimuth=None, tracking_backtracking_type=None,
                     minimum_tracking_limit_angle_d=-60.0, maximum_tracking_limit_angle_d=60.0,
                     lateral_intermodule_gap=0.02, vertical_intermodule_gap=0.02, table_to_table_spacing=0.0,
                     module_quality=None, module_mismatch_coefficient=None, light_induced_degradation=None,
                     dc_wiring_loss_at_stc=1.5, dc_health=1.0, heat_balance_conductive_coef=None,
                     heat_balance_convective_coef=None, sandia_conductive_coef=None, sandia_convective_coef=None,
                     cell_to_module_temp_diff=None, tracker_load_loss=0.0, post_height=None, structure_shading=0.0,
                     backside_mismatch=None):
        """
        A "power plant builder" helper method that adds a DC field to an inverter specified by :py:data:`inverter_name`,
        which is a child of the array  :py:data:`array_name`, which is a child of a block specified by
        :py:data:`block_name` on the :py:class:`~plantpredict.powerplant.PowerPlant`. DC field naming is sequential
        (numerically) - for instance, if there are 2 existing DC fields with names :py:data:`1` and :py:data:`2`
        (accessible via key :py:data:`name` for a given DC field dictionary), the next array created by
        :py:meth:`~plantpredict.powerplant.PowerPlant.add_dc_field` will automatically have :py:data:`name` equal to
        :py:data:`3`. This method does not currently account for the situation in which an existing power plant has
        DC fields named non-sequentially.

        Note that this addition is not persisted to PlantPredict unless
        :py:meth:`~plantpredict.powerplant.PowerPlant.update` is subsequently called.

        :param int block_name: Name (1-indexed integer) of the parent block to add DC field to. Can be found in the
                               relevant block dictionary (in attribute :py:attr:`blocks`) with key :py:data:`id`. This
                               value is returned for a new block when you create one with
                               :py:meth:`~plantpredict.powerplant.PowerPlant.add_block`. Must be between :py:data:`1`
                               and :py:data:`99`.
        :param int array_name: Name (1-indexed integer) of the parent array to add DC field to. This value is
                               returned for a new array when you create one with
                               :py:meth:`~plantpredict.powerplant.PowerPlant.add_array`. Must be between :py:data:`1`
                               and :py:data:`99`.
        :param str inverter_name: Name (letter) of the parent array to add the DC field to. This value is returned for
                                  a new array when you create one with
                                  :py:meth:`~plantpredict.powerplant.PowerPlant.add_inverter`. Must be only 1
                                  character.
        :param int module_id: Unique identifier of the module to be used in the DC field.
        :param int tracking_type: Represents the tracking type/mounting structure (Fixed Tilt or Tracker) of the DC
                                  field. Use :py:class:`~plantpredict.enumerations.TrackingTypeEnum`. (Seasonal Tilt
                                  currently not supported in this package).
        :param int modules_high: Number of modules high per table (number of ranks). Must be between :py:data:`1` and
                                 :py:data:`50`.
        :param int modules_wired_in_series: The number of modules electrically connected in series in a string.
        :param float post_to_post_spacing: Row spacing. Must be between :py:data:`0.0` and :py:data:`50.0` - units
                                           :py:data:`[m]`.
        :param number_of_rows: Number of rows of tables in DC field. Must be between :py:data:`1` and
                               :py:data:`10000`. Defaults to :py:data:`1`.
        :type number_of_rows: int, None
        :param int strings_wide: Number of strings across per table. Multiplied by :py:data:`modules_wired_in_series` to
                                 determine :py:attr:`modules_wide`. Must result in :py:attr:`modules_wide` between
                                 :py:data:`1` and  :py:data:`100`. Defaults to :py:data:`1`.
        :param field_dc_power: DC capacity of the DC field. Defaults to `None`. Non-null value required if
                               :py:data:`number_of_series_strings_wired_in_parallel` is `None` and  must be between
                               :py:data:`1` and :py:data:`20000` - units :py:data:`[kW]`.
        :type field_dc_power: float, None
        :param number_of_series_strings_wired_in_parallel: Number of strings of modules electrically connected in
                                                           parallel in the DC field. Defaults to :py:data:`None`.
                                                           Non-null value required if :py:data:`field_dc_power` is
                                                           :py:data:`None`, and must be between :py:data:`1` and
                                                           :py:data`10000`.
        :type number_of_series_strings_wired_in_parallel: float, None
        :param module_tilt: Tilt angle of modules in DC Field for a fixed tilt array. Defaults to :py:data:`None`.
                            Non-null value required required if :py:data:`tracking_type` is equal to
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.FIXED_TILT`, and must be between
                            :py:data:`0` and :py:data:`90` - units :py:data:`[degrees]`.
        :type module_tilt: float, None
        :param module_orientation: Represents the orientation (portrait or landscape) of modules in the DC field. If
                                   left as default (:py:data:`None`), is automatically set as the
                                   :py:attr:`module_orientation` of the module model specified by :py:data:`module_id`.
                                   Use :py:class:`~plantpredict.enumerations.ModuleOrientationEnum`.
        :type module_orientation: int, None
        :param module_azimuth: Orientation of the entire DC field. The convention is :py:data:`0.0` degrees for
                               North-facing arrays. If left as default (:py:data:`None`), is set to :py:data:`180.0`.
                               Must be between :py:data:`0` and :py:data:`360` - units :py:data:`[degrees]`.
        :type module_azimuth: float, None
        :param tracking_backtracking_type: Represents the backtracking algorithm (True-Tracking or Backtracking) used in
                                           DC Field. Use :py:class:`~plantpredict.enumerations.BacktrackingTypeEnum`.
        :type tracking_backtracking_type: int, None
        :param float minimum_tracking_limit_angle_d: Minimum tracking angle for horizontal tracker array. Defaults to
                                                     :py:data:`-60.0`. Must be between :py:data:`-90` and :py:data:`0` -
                                                     units :py:data:`[degrees]`.
        :param float maximum_tracking_limit_angle_d: Maximum tracking angle for horizontal tracker array. Defaults to
                                                     :py:data:`60.0`. Must be between :py:data:`0` and :py:data:`90` -
                                                     units :py:data:`[degrees]`.
        :param float lateral_intermodule_gap: Lateral gap between each module on the mounting structure. Defaults to
                                              :py:data:`0.02`. Must be between :py:data:`0` and py:data:`1` - units :py:data:`[m]'.
        :param float vertical_intermodule_gap: Vertical gap between each module on the mounting structure. Defaults to
                                               :py:data:`0.02`. Must be between :py:data:`0` and py:data:`1` - units
                                               :py:data:`[m]'.
        :param float table_to_table_spacing: Space between tables in each row. Defaults to :py:data:`0.0`. Must be
                                             between :py:data:`0` and :py:data:`50`.
        :param module_quality: Accounts for any discrepancy between manufacturer nameplate rating of module and actual
                               performance. If left as default (:py:data:`None`), is automatically set as the
                               :py:attr:`module_quality` of the module model specified by :py:data:`module_id`. Must be
                               between :py:data:`-200` and :py:data:`99` - units :py:data:`[%]`.
        :type module_quality: float, None
        :param module_mismatch_coefficient: Accounts for losses due to mismatch in electrical characteristics among
                                            modules in the strings of the DC fields (and between strings in the DC
                                            field). If left as default (:py:data:`None`), is automatically set as the
                                            :py:attr:`module_mismatch_coefficient` of the module model specified by
                                            :py:data:`module_id`. Must be between :py:data:`0` and :py:data:`30` -
                                            units :py:data:`[%]`.
        :type module_mismatch_coefficient: float, None
        :param light_induced_degradation: Accounts for losses due to light induced degradation. If left as default
                                          (:py:data:`None`), is automatically set as the
                                          :py:attr:`light_induced_degradation` of the module model specified by
                                          :py:data:`module_id`. Must be between :py:data:`0` and :py:data:`30` - units
                                          :py:data:`[%]`.
        :type light_induced_degradation: float, None
        :param float dc_wiring_loss_at_stc: Accounts for losses across all electrical wiring in the DC field. Defaults
                                            to :py:data:`1.5`. Must be between :py:data:`0` and :py:data:`30` - units
                                            :py:data:`[%]`.
        :param float dc_health: Accounts for any losses related to DC health. Defaults to :py:data:`1.0`. Must be
                                between :py:data:`-10` and :py:data:`10` - units :py:data:`[%]`.
        :param heat_balance_conductive_coef: Thermal loss factor (constant component) of heat balance module surface
                                             temperature model. If left as default (:py:data:`None`), is automatically
                                             set as the :py:attr:`heat_balance_conductive_coef` of the module model
                                             specified by :py:data:`module_id`. Must be between :py:data:`0` and
                                             :py:data:`100`. This value is only used if :py:attr:`model_temp_model` is
                                             set to
                                             :py:attr:`~plantpredict.enumerations.ModuleTemperatureModelEnum.HEAT_BALANCE`
                                             for the :py:class:`~plantpredict.prediction.Prediction` associated with the
                                             power plant by the attributes :py:attr:`project_id` and
                                             :py:attr:`prediction_id`.
        :type heat_balance_conductive_coef: float, None
        :param heat_balance_convective_coef: Thermal loss factor (wind speed component) of heat balance module surface
                                             temperature model. If left as default (:py:data:`None`), is automatically
                                             set as the :py:attr:`heat_balance_convective_coef` of the module model
                                             specified by :py:data:`module_id`. Must be between :py:data:`0` and
                                             :py:data:`100`. This value is only used if :py:attr:`model_temp_model` is
                                             set to
                                             :py:attr:`~plantpredict.enumerations.ModuleTemperatureModelEnum.HEAT_BALANCE`
                                             for the :py:class:`~plantpredict.prediction.Prediction` associated with the
                                             power plant by the attributes :py:attr:`project_id` and
                                             :py:attr:`prediction_id`.
        :type heat_balance_convective_coef: float, None
        :param sandia_conductive_coef: Coefficient :py:data:`a` for the Sandia module surface temperature model. If left
                                       as default (:py:data:`None`), is automatically set as the
                                       :py:attr:`sandia_conductive_coef` of the module model specified by
                                       :py:data:`module_id`. Must be between :py:data:`-5` and :py:data:`0`. This value
                                       is only used if :py:attr:`model_temp_model` is set to
                                       :py:attr:`~plantpredict.enumerations.ModuleTemperatureModelEnum.SANDIA` for the
                                       :py:class:`~plantpredict.prediction.Prediction` associated with the power plant
                                       by attributes :py:attr:`project_id` and :py:attr:`prediction_id`.
        :type sandia_conductive_coef: float, None
        :param sandia_convective_coef: Coefficient :py:data:`b` for the Sandia module surface temperature model. If left
                                       as default (:py:data:`None`), is automatically set as the
                                       :py:attr:`sandia_convective_coef` of the module model specified by
                                       :py:data:`module_id`. Must be between :py:data:`-1` and :py:data:`0`. This value
                                       is only used if :py:attr:`model_temp_model` is set to
                                       :py:attr:`~plantpredict.enumerations.ModuleTemperatureModelEnum.SANDIA` for the
                                       :py:class:`~plantpredict.prediction.Prediction` associated with the power plant
                                       by attributes :py:attr:`project_id` and :py:attr:`prediction_id`.
        :type sandia_convective_coef: float, None
        :param cell_to_module_temp_diff: Difference between surface and cell temperature of modules. If left as default
                                         (:py:data:`None`), is automatically set as the
                                         :py:attr:`cell_to_module_temp_diff` of the module model specified by
                                         :py:data:`module_id`. Must be between :py:data:`0` and :py:data:`15` - units
                                         :py:data:`[degrees-C]`.
        :type cell_to_module_temp_diff: float, None
        :param float tracker_load_loss: Accounts for losses from power use of horizontal tracker system. Defaults to
                                        :py:data:`0.0`. Must be between :py:data:`0` and :py:data:`100` - units
                                        :py:data:`[%]`.
        :param post_height: Height of mounting structure (table) post. Defaults to :py:data:`None`. If left as default
                            (:py:data:`None`), automatically calculated as
                            :py:data:`((collector_bandwidth * sin(tilt) / 2) + 1`, where :py:data:`tilt` is
                            :py:data:`module_tilt` if :py:data:`tracking_type` is
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.FIXED_TILT`, or the largest of the
                            absolute values of
                            :py:data:`maximum_tracking_limit_angle_d`/:py:data:`minimum_tracking_limit_angle_d`
                            if :py:data:`tracking_type` is
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.HORIZONTAL_TRACKER`. However, if the
                            calculated value is less than :py:data:`1.5`, :py:data:`post_height` is defaulted to
                            :py:data:`1.5`. Must be between :py:data:`0` and :py:data:`50` - units :py:data:`[m]`. This
                            value is only used if the module model specified with :py:data:`module_id` is bifacial.
        :type post_height: float, None
        :param float structure_shading: Accounts for backside of module losses from structure shading. Defaults to
                                        :py:data:`0.0`. Must be between :py:data:`0` and :py:data:`100` - units
                                        :py:data:`[%]`. This value is only used if the module model specified with
                                        :py:data:`module_id` is bifacial.
        :param backside_mismatch: Accounts for losses due to inconsistent backside irradiance among modules in the DC
                                  field. Defaults to :py:data:`None`. If left as default (:py:data:`None`), is
                                  automatically set as the :py:attr:`module_orientation` of the module model specified
                                  by :py:data:`module_id`. Must be between :py:data:`0` and :py:data:`100` - units
                                  :py:data:`[%]`. This value is only used if the module model specified with
                                  :py:data:`module_id` is bifacial.
        :type backside_mismatch: float, None
        :raises ValueError: Raised if :py:data:`block_name` is not a valid block name in the existing power plant, or
                            if the :py:data:`block_name` is valid but :py:data:`array_name` is not a valid array name
                            in the block, or if :py:data:`array_name` is valid but :py:data:`inverter_name` is not a
                            valid inverter in the array. Also raised if :py:data:`tracking_type` is
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.FIXED_TILT` and :py:data:`module_tilt`
                            is :py:data:`None`, or if :py:data:`tracking_type` is
                            :py:attr:`~plantpredict.enumerations.TrackingTypeEnum.HORIZONTAL_TRACKER` and
                            :py:data:`tracking_backtracking_type` is :py:data:`None`. Also raised if both
                            :py:data:`field_dc_power` and :py:data`number_of_series_strings_wired_in_parallel` are
                            :py:data:`None` or are both not :py:data:`None`. Also raised if :py:data:`tracking_type` is
                            :py:data:`~plantpredict.enumerations.TrackingTypeEnum.SEASONAL_TILT`.
        :return: The name of the newly added DC field.
        :rtype: int
        """
        # only fixed tilt and horizontal tracker supported
        if tracking_type == TrackingTypeEnum.SEASONAL_TILT:
            raise ValueError("Seasonal Tilt is not currently supported by the add_dc_field method.")

        # validate inputs
        self._validate_inverter_name(block_name=block_name, array_name=array_name, inverter_name=inverter_name)
        self._validate_mounting_structure_parameters(tracking_type, module_tilt, tracking_backtracking_type)

        # calculate parameters typically calculated in the UI
        m = self.api.module(id=module_id)
        m.get()
        field_dc_power, number_of_series_strings_wired_in_parallel = self._validate_dc_field_sizing(
            field_dc_power=field_dc_power,
            number_of_series_strings_wired_in_parallel=number_of_series_strings_wired_in_parallel,
            planned_module_rating=m.stc_max_power,
            modules_wired_in_series=modules_wired_in_series,
        )
        module_orientation = module_orientation if module_orientation is not None else m.default_orientation
        modules_wide = self._calculate_modules_wide(strings_wide, modules_wired_in_series)
        collector_bandwidth = self._calculate_collector_bandwidth(
            module_width=m.width,
            module_length=m.length,
            module_orientation=module_orientation,
            vertical_intermodule_gap=vertical_intermodule_gap,
            modules_high=modules_high
        )
        table_length = self._calculate_table_length(
            modules_wide=modules_wide,
            module_width=m.width,
            module_length=m.length,
            module_orientation=module_orientation,
            lateral_intermodule_gap=lateral_intermodule_gap
        )
        tables_per_row = self._calculate_tables_per_row(
            field_dc_power=field_dc_power,
            planned_module_rating=m.stc_max_power,
            modules_high=modules_high,
            modules_wide=modules_wide,
            number_of_rows=number_of_rows
        )
        self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][ord(inverter_name) - 65]["dc_fields"].append(
            {
                "name": len(self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][
                                ord(inverter_name) - 65]["dc_fields"]) + 1,
                "module_id": module_id,
                "tracking_type": tracking_type,
                "minimum_tracking_limit_angle_d": minimum_tracking_limit_angle_d,
                "maximum_tracking_limit_angle_d": maximum_tracking_limit_angle_d,
                "module_orientation": module_orientation,
                "modules_high": modules_high,
                "module_azimuth": (module_azimuth if module_azimuth is not None
                                   else self._get_default_module_azimuth_from_latitude()),
                "collector_bandwidth": collector_bandwidth,
                "post_to_post_spacing": post_to_post_spacing,
                # Electrical
                "planned_module_rating": m.stc_max_power,
                "modules_wired_in_series": modules_wired_in_series,
                "field_dc_power": field_dc_power,
                "number_of_series_strings_wired_in_parallel": number_of_series_strings_wired_in_parallel,
                "module_count": 1000*field_dc_power/m.stc_max_power,    # confirmed calculation in PlantPredict backend
                # Losses
                "module_quality": module_quality if module_quality is not None else m.module_quality,
                "module_mismatch_coefficient": (module_mismatch_coefficient if module_mismatch_coefficient is not None
                                                else m.module_mismatch_coefficient),
                "light_induced_degradation": (light_induced_degradation if light_induced_degradation is not None else
                                              m.light_induced_degradation),
                "dc_wiring_loss_at_stc": dc_wiring_loss_at_stc,
                "dc_health": dc_health,
                "heat_balance_conductive_coef": (heat_balance_conductive_coef if heat_balance_conductive_coef is not
                                                 None else m.heat_balance_conductive_coef),
                "heat_balance_convective_coef": (heat_balance_convective_coef if heat_balance_convective_coef is not
                                                 None else m.heat_balance_convective_coef),
                "sandia_conductive_coef": (sandia_conductive_coef if sandia_conductive_coef is not None
                                           else m.sandia_conductive_coef),
                "cell_to_module_temp_diff": (cell_to_module_temp_diff if cell_to_module_temp_diff is not None else
                                             m.cell_to_module_temp_diff),
                "sandia_convective_coef": (sandia_convective_coef if sandia_convective_coef is not None
                                           else m.sandia_convective_coef),
                "tracker_load_loss": tracker_load_loss,
                # Advanced Fields
                "lateral_intermodule_gap": lateral_intermodule_gap,
                "vertical_intermodule_gap": vertical_intermodule_gap,
                "modules_wide": modules_wide,
                "table_to_table_spacing": table_to_table_spacing,
                "number_of_rows": number_of_rows,
                "table_length": table_length,
                "tables_per_row": tables_per_row,
                "field_length": self._calculate_dc_field_length(tables_per_row, module_orientation, m.length, m.width,
                                                                lateral_intermodule_gap, modules_wide, tracking_type,
                                                                number_of_rows, post_to_post_spacing,
                                                                collector_bandwidth),
                "field_width": self._calculate_dc_field_width(tracking_type, number_of_rows, post_to_post_spacing,
                                                              collector_bandwidth, tables_per_row, module_orientation,
                                                              m.length, m.width, lateral_intermodule_gap, modules_wide),
                "post_height": (post_height if post_height is not None
                                else self._calculate_default_post_height(tracking_type, collector_bandwidth,
                                                                         module_tilt, minimum_tracking_limit_angle_d,
                                                                         maximum_tracking_limit_angle_d)),
            }
        )

        # add module tilt if fixed tilt
        if tracking_type == TrackingTypeEnum.FIXED_TILT:
            self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][ord(inverter_name) - 65][
                "dc_fields"][-1].update({"module_tilt": module_tilt})

        # add backtracking type if horizontal tracker
        if tracking_type == TrackingTypeEnum.HORIZONTAL_TRACKER:
            self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][ord(inverter_name) - 65][
                "dc_fields"][-1].update({"tracking_backtracking_type": tracking_backtracking_type})

        # add bifacial parameters if module is bifacial
        if m.faciality == FacialityEnum.BIFACIAL:
            self.blocks[block_name - 1]["arrays"][array_name - 1]["inverters"][ord(inverter_name) - 65][
                "dc_fields"][-1].update({
                    "structure_shading": structure_shading,
                    "backside_mismatch": backside_mismatch if backside_mismatch is not None else m.backside_mismatch
                })

        return self.blocks[
            block_name - 1]["arrays"][array_name - 1]["inverters"][ord(inverter_name) - 65]["dc_fields"][-1]["name"]

    def __init__(self, api, project_id=None, prediction_id=None, use_cooling_temp=True, **kwargs):
        """
        Constructor method.
        """
        self.project_id = project_id
        self.prediction_id = prediction_id
        self.use_cooling_temp = use_cooling_temp

        self.power_factor = 1.0
        self.blocks = []
        self.transformers = []
        self.transmission_lines = []

        # set any provided keyword arguments as attributes
        self.__dict__.update(kwargs)

        super(PowerPlant, self).__init__(api)



import unittest
import mock

import plantpredict
from plantpredict import project, prediction, powerplant, geo, inverter, module, weather, ashrae
from tests import mocked_requests


class TestApi(unittest.TestCase):
    @mock.patch('plantpredict.api.requests.post', new=mocked_requests.mocked_requests_post)
    def setUp(self):
        self.api = plantpredict.Api(
            client_id="dummy client id",
            client_secret="dummy client secret"
        )

    @mock.patch('plantpredict.api.requests.post', new=mocked_requests.mocked_requests_post)
    def test_init(self):
        self.assertEqual(self.api.base_url, "https://api.plantpredict.terabase.energy")
        self.assertEqual(self.api.client_id, "dummy client id")
        self.assertEqual(self.api.client_secret, "dummy client secret")
        self.assertEqual(self.api.access_token, "dummy access token")
        self.assertEqual(self.api.refresh_token, "dummy refresh token")

    def test_project(self):
        self.assertIsInstance(self.api.project(), project.Project)

    def test_prediction(self):
        self.assertIsInstance(self.api.prediction(), prediction.Prediction)

    def test_powerplant(self):
        self.assertIsInstance(self.api.powerplant(), powerplant.PowerPlant)

    def test_geo(self):
        self.assertIsInstance(self.api.geo(), geo.Geo)

    def test_inverter(self):
        self.assertIsInstance(self.api.inverter(), inverter.Inverter)

    def test_module(self):
        self.assertIsInstance(self.api.module(), module.Module)

    def test_weather(self):
        self.assertIsInstance(self.api.weather(), weather.Weather)

    def test_ashrae(self):
        self.assertIsInstance(self.api.ashrae(), ashrae.ASHRAE)


if __name__ == '__main__':
    unittest.main()



import requests
import json

from plantpredict.plant_predict_entity import PlantPredictEntity
from plantpredict.error_handlers import handle_refused_connection, handle_error_response
from plantpredict.utilities import convert_json, camel_to_snake


class Project(PlantPredictEntity):
    """
    The Project entity in PlantPredict defines the location info and serves as a container for any number of Predictions.
    """
    def create(self):
        """
        **POST** */Project/*

        Creates a new :py:mod:`plantpredict.Project` entity in the PlantPredict database using the attributes
        assigned to the local object instance. Automatically assigns the resulting :py:attr:`id` to the local object
        instance. See the minimum required attributes (below) necessary to successfully create a new
        :py:mod:`plantpredict.Project`. Note that the full scope of attributes is not limited to the minimum
        required set.

        Use :py:meth:`plantpredict.Project.assign_location_attributes` to automatically assign all required (and
        non-required) location-related/geological attributes.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes for successful Prediction creation
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    name; str; Name of the project
                    latitude; float; North-South GPS coordinate of the Project location. Must be between :py:data:`-90` and :py:data:`90` - units :py:data:`[decimal degrees]`.
                    longitude; float; East-West coordinate of the Project location, in decimal degrees. Must be between :py:data:`-180` and :py:data:`180` units :py:data:`[decimal degrees]`.
                    country; str; Full name of the country of the Project's location.
                    country_code; str; Country code of the Project's location (ex. US for United States, AUS for Australia, etc.)
                    elevation; float; The elevation of the Project location above seal level units :py:data:`[m]`.
                    standard_offset_from_utc; float; Time zone with respect to Greenwich Mean Time (GMT) in +/- hours offset.
        """
        self.create_url_suffix = "/Project"

        return super(Project, self).create()

    def delete(self):
        """HTTP Request: DELETE /Project/{ProjectId}

        Deletes an existing Project entity in PlantPredict. The local instance of the Project entity must have
        attribute self.id identical to the project id of the Project to be deleted.

        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """

        self.delete_url_suffix = "/Project/{}".format(self.id)

        return super(Project, self).delete()

    def get(self):
        """HTTP Request: GET /Project/{Id}

        Retrieves an existing Project entity in PlantPredict and automatically assigns all of its attributes to the
        local Project object instance. The local instance of the Project entity must have attribute self.id identical
        to the project id of the Project to be retrieved.

        :return: A dictionary containing all of the retrieved Project attributes.
        :rtype: dict
        """

        self.get_url_suffix = "/Project/{}".format(self.id)

        return super(Project, self).get()

    def update(self):
        """HTTP Request: PUT /Project/{ProjectId}

        Updates an existing Project entity in PlantPredict using the full attributes of the local Project instance.
        Calling this method is most commonly preceded by instantiating a local instance of Project with a specified
        project id, calling the Project.get() method, and changing any attributes locally.

        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """
        self.update_url_suffix = "/Project/{}".format(self.id)

        return super(Project, self).update()

    @handle_refused_connection
    @handle_error_response
    def get_all_predictions(self):
        """HTTP Request: GET /Project/{ProjectId}/Prediction

        Retrieves the full attributes of every Prediction associated with the Project.

        :return: A list of dictionaries, each containing the attributes of a Prediction entity.
        :rtype: list of dict
        """

        return requests.get(
            url=self.api.base_url + "/Project/{}/Prediction".format(self.id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

    def search(self, latitude, longitude, search_radius=1.0):
        """HTTP Request: GET /Project/Search

        Searches for all existing Project entities within a search radius of a specified latitude/longitude.

        :param latitude: North-South coordinate of the Project location, in decimal degrees.
        :type latitude: float
        :param longitude: East-West coordinate of the Project location, in decimal degrees.
        :type longitude: float
        :param search_radius: search radius in miles
        :type search_radius: float
        :return: TODO
        """
        response = requests.get(
            url=self.api.base_url + "/Project/Search",
            headers={"Authorization": "Bearer " + self.api.access_token},
            params={'latitude': latitude, 'longitude': longitude, 'searchRadius': search_radius}
        )

        project_list = response.json()

        return [convert_json(p, camel_to_snake) for p in project_list]

    @handle_refused_connection
    @handle_error_response
    def assign_location_attributes(self):
        """

        :return:
        """
        geo = self.api.geo(latitude=self.latitude, longitude=self.longitude)
        geo.get_location_info()
        geo.get_elevation()
        geo.get_time_zone()

        self.locality = geo.locality
        self.state_province_code = geo.state_province_code
        self.state_province = geo.state_province
        self.country_code = geo.country_code
        self.country = geo.country
        self.region = geo.region
        self.elevation = geo.elevation
        self.standard_offset_from_utc = geo.time_zone

    def __init__(self, api, id=None, name=None, latitude=None, longitude=None):
        if id:
            self.id = id
        self.name = name
        self.latitude = latitude
        self.longitude = longitude

        self.locality = None
        self.state_province_code = None
        self.state_province = None
        self.country_code = None
        self.country = None
        self.region = None
        self.elevation = None
        self.standard_offset_from_utc = None

        super(Project, self).__init__(api)



import json
import requests
from plantpredict.utilities import convert_json, camel_to_snake, decorate_all_methods
from plantpredict.error_handlers import handle_refused_connection, handle_error_response


@decorate_all_methods(handle_refused_connection)
@decorate_all_methods(handle_error_response)
class Geo(object):
    """
    The :py:mod:`Geo` entity is used to get location-related information for a given latitude/longitude. Its methods can
    be used individually, but typically location related info is needed in the context of a PlantPredict
    :py:mod:`Project` entity. In this case the user can simply call the method :py:meth:`Project.get_location_info`
    which calls all :py:mod:`Geo` class methods and automatically assigns all location-related attributes to that
    instance of :py:mod:`Project`. Note: This API resource does not represent a database entity in PlantPredict. This
    is a simplified connection to the Google Maps API. See Google Maps API Reference for further functionality.
    (https://developers.google.com/maps/)
    """

    def get_location_info(self):
        """
        **GET** */Geo/* :py:attr:`latitude` */* :py:attr:`longitude` */Location*

        Retrieves pertinent location info for a given latitude and longitude such as locality, state/province, country,
        etc. In addition to returning a dictionary with this information, the method also automatically assigns the
        contents of the dictionary to the instance of :py:mod:`Geo` as attributes.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes on object to call this method successfully.
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    latitude; float; North-South GPS coordinate. Must be between :py:data:`-90` and :py:data:`90` - units :py:data:`[decimal degrees]`.
                    longitude; float; East-West GPS coordinate Must be between :py:data:`-180` and :py:data:`180` units :py:data:`[decimal degrees]`.

        .. container:: toggle

            .. container:: header

                **Example Code**

            .. container:: example_code

                Instantiate a local object instance of :py:mod:`Geo` with latitude and longitude as inputs (which
                automatically assigns them as attributes to the object). Then call the method on the object.

                .. code-block:: python

                    geo = api.geo(latitude=35.1, longitude=-106.7)
                    geo.get_location_info()

        .. container:: toggle

            .. container:: header

                **Example Response**

            .. container:: example_response

                The method returns a dictionary as shown in the example below, and assigns its contents as attributes
                to the local object instance of :py:mod:`Geo`.

                .. code-block:: python

                    {
                        "country": "United States",
                        "country_code": "US",
                        "locality": "Albuquerque",
                        "region": "North America",
                        "state_province": "New Mexico",
                        "state_province_code": "NM
                    }

        :return: A dictionary with location information as shown in "Example Response".
        :rtype: dict
        """
        response = requests.get(
            url=self.api.base_url + "/Geo/{}/{}/Location".format(self.latitude, self.longitude),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def get_elevation(self):
        """
        **GET** */Geo/* :py:attr:`latitude` */* :py:attr:`longitude` */Elevation*

        Retrieves the elevation in meters for a given latitude and longitude. In addition to returning a dictionary with
        this information, the method also automatically assigns the  contents of the dictionary to the instance of
        :py:mod:`Geo` as attributes.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes on object to call this method successfully.
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    latitude; float; North-South GPS coordinate. Must be between :py:data:`-90` and :py:data:`90` - units :py:data:`[decimal degrees]`.
                    longitude; float; East-West GPS coordinate Must be between :py:data:`-180` and :py:data:`180` units :py:data:`[decimal degrees]`.

        .. container:: toggle

            .. container:: header

                **Example Code**

            .. container:: example_code

                Instantiate a local object instance of :py:mod:`Geo` with latitude and longitude as inputs (which
                automatically assigns them as attributes to the object). Then call the method on the object.

                .. code-block:: python

                    geo = api.geo(latitude=35.1, longitude=-106.7)
                    geo.get_elevation()

        .. container:: toggle

            .. container:: header

                **Example Response**

            .. container:: example_response

                The method returns a dictionary as shown in the example below, and assigns its contents as attributes
                to the local object instance of :py:mod:`Geo`.

                .. code-block:: python

                    {
                        "elevation": 1553.614
                    }

        :return: A dictionary with location information as shown in "Example Response".
        :rtype: dict
        """
        response = requests.get(
            url=self.api.base_url + "/Geo/{}/{}/Elevation".format(self.latitude, self.longitude),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def get_time_zone(self):
        """
        **GET** */Geo/* :py:attr:`latitude` */* :py:attr:`longitude` */TimeZone*

        Retrieves the time zone as a time shift in hours with respect to GMT for a given latitude and longitude. In
        addition to returning a dictionary with this information, the method also automatically assigns the  contents
        of the dictionary to the instance of :py:mod:`Geo` as attributes.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes on object to call this method successfully.
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    latitude; float; North-South GPS coordinate. Must be between :py:data:`-90` and :py:data:`90` - units :py:data:`[decimal degrees]`.
                    longitude; float; East-West GPS coordinate Must be between :py:data:`-180` and :py:data:`180` units :py:data:`[decimal degrees]`.

        .. container:: toggle

            .. container:: header

                **Example Code**

            .. container:: example_code

                Instantiate a local object instance of :py:mod:`Geo` with latitude and longitude as inputs (which
                automatically assigns them as attributes to the object). Then call the method on the object.

                .. code-block:: python

                    geo = api.geo(latitude=35.1, longitude=-106.7)
                    geo.get_time_zone()

        .. container:: toggle

            .. container:: header

                **Example Response**

            .. container:: example_response

                The method returns a dictionary as shown in the example below, and assigns its contents as attributes
                to the local object instance of :py:mod:`Geo`.

                .. code-block:: python

                    {
                        "time_zone": -7.0
                    }

        :return: A dictionary with location information as shown in "Example Response".
        :rtype: dict
        """
        response = requests.get(
            url=self.api.base_url + "/Geo/{}/{}/TimeZone".format(self.latitude, self.longitude),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def __init__(self, api, latitude=None, longitude=None):
        """
        Initializes a local object instance of :py:mod:`Geo`.

        :param api: :py:class:`~plantpredict.api.Api` object.
        :type api: object
        :param latitude: North-South GPS coordinate. Must be between :py:data:`-90` and :py:data:`90` - units :py:data:`[decimal degrees]`.
        :type latitude: float
        :param longitude: East-West GPS coordinate Must be between :py:data:`-180` and :py:data:`180` units :py:data:`[decimal degrees]`.
        :type longitude: float
        """
        self.api = api

        self.latitude = latitude
        self.longitude = longitude

        super(Geo, self).__init__()



import json
import requests

from plantpredict.utilities import convert_json, camel_to_snake, snake_to_camel, decorate_all_methods
from plantpredict.error_handlers import handle_refused_connection, handle_error_response, APIError


@decorate_all_methods(handle_refused_connection)
@decorate_all_methods(handle_error_response)
class PlantPredictEntity(object):
    def create(self, *args):
        """Generic POST request."""
        response = requests.post(
            url=self.api.base_url + self.create_url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=convert_json(self.__dict__, snake_to_camel)
        )

        # power plant is the exception that doesn't have its own id. has a project and prediction id
        try:
            self.id = response.json()['id'] if 200 <= response.status_code < 300 else None
        except ValueError:
            pass

        return response

    def delete(self):
        """Generic DELETE request."""

        return requests.delete(
            url=self.api.base_url + self.delete_url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

    def get(self):
        """Generic GET request."""
        response = requests.get(
            url=self.api.base_url + self.get_url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        if response.status_code == 404:
            raise APIError(response.status_code, response.content, response.url)
        else:
            attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def update(self):
        """Generic PUT request."""

        return requests.put(
            url=self.api.base_url + self.update_url_suffix,
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=convert_json(self.__dict__, snake_to_camel)
        )

    def __init__(self, api, **kwargs):
        self.api = api

        self.create_url_suffix = None
        self.delete_url_suffix = None
        self.get_url_suffix = None
        self.update_url_suffix = None

        self.__dict__.update(kwargs)



import mock
import json
import unittest

from plantpredict.ashrae import ASHRAE
from tests import plantpredict_unit_test_case, mocked_requests


class TestPrediction(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.ashrae.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_station(self):
        self._make_mocked_api()
        ashrae = ASHRAE(api=self.mocked_api, latitude=35.0, longitude=-109.0)
        response = ashrae.get_station(station_name="TEST STATION")

        self.assertEqual(response.json(), {
                "station_name": "TEST STATION",
                "wmo": 18081,
                "cool_996": 20.0,
                "min_50_year": -20.0,
                "max_50_year": 17.0,
                "distance": 5.3,
                "latitude": 35.0,
                "longitude": -109.0
        })
        self.assertEqual(ashrae.cool_996, 20.0)

    @mock.patch('plantpredict.ashrae.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_closest_station(self):
        self._make_mocked_api()
        ashrae = ASHRAE(api=self.mocked_api, latitude=33.0, longitude=-110.0)
        response = ashrae.get_closest_station()

        self.assertEqual(response.json(), {
                "station_name": "TEST STATION",
                "wmo": 18081,
                "cool_996": 20.0,
                "min_50_year": -20.0,
                "max_50_year": 17.0,
                "distance": 5.3,
                "latitude": 35.0,
                "longitude": -109.0
        })
        self.assertEqual(ashrae.cool_996, 20.0)




import pandas as pd


def load_from_excel(file_path, sheet_name=None):
    """
    Loads the data from an Excel file into a list of dictionaries, where each dictionary represents a row in the Excel
    file and the keys of each dictionary represent each column header in the Excel file. The method creates this list
    of dictionaries via a Pandas dataframe.

    :param file_path: The full file path (appended with .xlsx) of the Excel file to be loaded.
    :type file_path: str
    :param sheet_name: Name of a particular sheet in the file to load (optional, defaults to the first sheet in the
    Excel file).
    :type sheet_name: str
    :return: List of dictionaries, each dictionary representing a row in the Excel file.
    :rtype: list of dict
    """
    xl = pd.ExcelFile(file_path)
    sheet_name = sheet_name if sheet_name else xl.sheet_names[0]

    return xl.parse(sheet_name, index_col=None).to_dict('records')


def export_to_excel(data, file_path, sheet_name="Sheet1", field_order=None, sorting_fields=None):
    """
    Writes data from a list of dictionaries to an Excel file, where each dictionary represents a row in the Excel file
    and the keys of each dictionary represent each column header in the Excel file.

    :param data: List of dictionaries, each dictionary representing a row in the Excel file.
    :type data: list of dict
    :param file_path: The full file path (appended with .xlsx) of the Excel file to be written to. This will overwrite
    data if both file_path and sheet_name already exist.
    :type file_path: str
    :param sheet_name: Name of a particular sheet in the file to write to (optional, defaults to "Sheet1").
    :type sheet_name: str
    :param field_order: List of keys from data ordered to match the intended Excel column ordering (left to right). Must
    include all keys/columns. Any keys omitted from the list will not be written as columns. (optional)
    :type field_order: list of str
    :param sorting_fields: List of keys from data to be used as sorting columns (small to large) in Excel. Can be any
    length from 1 column to every column. The order of the list will dictate the sorting order.
    :type sorting_fields: list of str
    :return: None
    """

    writer = pd.ExcelWriter(file_path, engine='openpyxl')
    df = pd.DataFrame(data)
    if field_order:
        df = df[field_order]
    if sorting_fields:
        df = df.sort_values(sorting_fields)
    df.to_excel(writer, sheet_name=sheet_name, index=False)
    writer.save()



import mock
import json
import unittest

from plantpredict.prediction import Prediction
from tests import plantpredict_unit_test_case, mocked_requests


class TestPrediction(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.create')
    def test_create(self, mocked_create):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=7)

        prediction.create(use_closest_ashrae_station=False)
        self.assertEqual(prediction.create_url_suffix, "/Project/7/Prediction")
        self.assertTrue(mocked_create.called)

    @mock.patch('plantpredict.ashrae.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_assign_plant_design_temperature_with_closest_ashrae_station(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=7)

        prediction._assign_plant_design_temperature_with_closest_ashrae_station()
        self.assertEqual(prediction.ashrae_station, "TEST STATION")
        self.assertEqual(prediction.cool_996, 20.0)
        self.assertEqual(prediction.max_50_year, 17.0)
        self.assertEqual(prediction.min_50_year, -20.0)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.delete')
    def test_delete(self, mocked_delete):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, id=77, project_id=7)

        prediction.delete()
        self.assertEqual(prediction.delete_url_suffix, "/Project/7/Prediction/77")
        self.assertTrue(mocked_delete.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.get')
    def test_get(self, mocked_get):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, id=77, project_id=7)

        prediction.get()
        self.assertEqual(prediction.get_url_suffix, "/Project/7/Prediction/77")
        self.assertTrue(mocked_get.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.update')
    def test_update(self, mocked_update):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=7)

        prediction.update()
        self.assertEqual(prediction.update_url_suffix, "/Project/7/Prediction")
        self.assertTrue(mocked_update.called)

    @mock.patch('plantpredict.prediction.Prediction._wait_for_prediction')
    @mock.patch('plantpredict.prediction.requests.post', new=mocked_requests.mocked_requests_post)
    def test_run(self, mocked_wait_for_prediction):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=710, id=555)

        is_success = prediction.run()
        self.assertTrue(mocked_wait_for_prediction.called)
        self.assertEqual(is_success["is_successful"], True)

    @mock.patch('plantpredict.prediction.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_results_summary(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=710, id=555)

        response = prediction.get_results_summary()
        self.assertEqual(response.json(), {
            "prediction_name": "Test Prediction", "block_result_summaries": [{"name": 1}]
        })

    @mock.patch('plantpredict.prediction.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_results_details(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=710, id=555)

        response = prediction.get_results_details()
        self.assertEqual(response.json(), {"prediction_name": "Test Prediction Details"})

    @mock.patch('plantpredict.prediction.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_nodal_data(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, project_id=710, id=555)

        nodal_data_dc_field = prediction.get_nodal_data(params={
            'block_number': 1,
            'array_number': 1,
            'inverter_name': 'A',
            'dc_field_number': 1
        })
        self.assertEqual(nodal_data_dc_field, {"nodal_data_dc_field": {}})

    @mock.patch('plantpredict.prediction.requests.post', new=mocked_requests.mocked_requests_post)
    @mock.patch('plantpredict.prediction.requests.get', new=mocked_requests.mocked_requests_get)
    def test_clone(self):
        self._make_mocked_api()

        prediction = Prediction(api=self.mocked_api, id=555, project_id=710)
        prediction.created_date = '2019-06-28 11:26:00'
        prediction.last_modified = '2019-06-28 11:26:00'
        prediction.last_modified_by = 'Stephen Kaplan'
        prediction.last_modified_by_id = 1
        prediction.project = {"id": 710}
        prediction.powerplant_id = 10101
        prediction.powerplant_id = {"powerplant_id": 10101}

        new_prediction_id = prediction.clone(new_prediction_name="Cloned Prediction")
        self.assertEqual(new_prediction_id, 556)

    def test_init_minimum_inputs(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api)

        self.assertEqual(prediction.api, self.mocked_api)
        with self.assertRaises(AttributeError):
            print(prediction.id)
        self.assertIsNone(prediction.project_id)
        self.assertIsNone(prediction.name)

        self.assertIsNone(prediction.status)
        self.assertIsNone(prediction.year_repeater)
        self.assertIsNone(prediction.error_spa_var)
        self.assertIsNone(prediction.error_model_acc)
        self.assertIsNone(prediction.error_int_ann_var)
        self.assertIsNone(prediction.error_sens_acc)
        self.assertIsNone(prediction.error_mon_acc)

    def test_init_maximum_inputs(self):
        self._make_mocked_api()
        prediction = Prediction(api=self.mocked_api, id=77, project_id=7, name="Test Prediction")
        self.assertEqual(prediction.api, self.mocked_api)
        self.assertEqual(prediction.id, 77)
        self.assertEqual(prediction.project_id, 7)
        self.assertEqual(prediction.name, "Test Prediction")

        self.assertIsNone(prediction.status)
        self.assertIsNone(prediction.year_repeater)
        self.assertIsNone(prediction.error_spa_var)
        self.assertIsNone(prediction.error_model_acc)
        self.assertIsNone(prediction.error_int_ann_var)
        self.assertIsNone(prediction.error_sens_acc)
        self.assertIsNone(prediction.error_mon_acc)


if __name__ == '__main__':
    unittest.main()



import requests
import json
from plantpredict.plant_predict_entity import PlantPredictEntity
from plantpredict.error_handlers import handle_refused_connection, handle_error_response, APIError
from plantpredict.enumerations import EntityTypeEnum


class Inverter(PlantPredictEntity):
    """
    """
    def create(self):
        """POST /Inverter"""
        self.create_url_suffix = "/Inverter"
        return super(Inverter, self).create()

    def delete(self):
        """DELETE /Inverter/{Id}"""
        self.delete_url_suffix = "/Inverter/{}".format(self.id)
        return super(Inverter, self).delete()

    def get(self):
        """GET /Inverter/{Id}"""
        self.get_url_suffix = "/Inverter/{}".format(self.id)
        return super(Inverter, self).get()

    def update(self):
        """PUT /Inverter"""
        self.update_url_suffix = "/Inverter".format(self.id)
        return super(Inverter, self).update()

    @handle_refused_connection
    @handle_error_response
    def upload_ond_file(self, file_name=None, file_path=None):
        """
        creates a new inverter from a source .ond file
        """
        json_parse = requests.post(
            url=self.api.base_url + "/Inverter/ParseONDFile",
            files=[('fileName', (file_name, open(file_path, 'rb'), 'application/octet-stream'))],
            headers={"Authorization": "Bearer " + self.api.access_token},
           )

        create_request = requests.post(
            url=self.api.base_url + "/Inverter",
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=json.loads(json_parse.content),
           )

        return json.loads(create_request.content)

    @handle_refused_connection
    @handle_error_response
    def parse_ond_file(self, file_name=None, file_path=None):
        """
        creates a new inverter from a source .ond file
        """
        json_parse = requests.post(
            url=self.api.base_url + "/Inverter/ParseONDFile",
            files=[('fileName', (file_name, open(file_path, 'rb'), 'application/octet-stream'))],
            headers={"Authorization": "Bearer " + self.api.access_token},
           )
        return json.loads(json_parse.content)

    @handle_refused_connection
    @handle_error_response
    def create_from_json(self, json_inverter=None):
        """
        creates a new inverter from a source JSON file
        """
        create_request = requests.post(
            url=self.api.base_url + "/Inverter",
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=json_inverter,
           )
        return json.loads(create_request.content)

    @handle_refused_connection
    @handle_error_response
    def change_status(self, new_status, note=""):
        """

        :param new_status:
        :param note:
        :return:
        """
        return requests.post(
            url=self.api.base_url + "/Inverter/Status",
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=[{
                "name": self.name,
                "id": self.id,
                "type": EntityTypeEnum.INVERTER,
                "status": new_status,
                "note": note
            }]
        )

    @handle_refused_connection
    @handle_error_response
    def get_kva(self, elevation, temperature, use_cooling_temp):
        """
        Uses the given elevation and temperature to interpolate a kVa rating from the inverter's kVa curves.

        :param float elevation: Elevation at which to evaluate the inverter kVa rating - units :py:data:`[m]`.
        :param float temperature: Temperature at which to evaluate the inverter kVa rating - units :py:data:`[deg-C]`.
        :param bool use_cooling_temp: Determines if the calculation should use the plant design cooling temperature (
                                      at 99.6 degrees).
        :return: # TODO after new API response is implemented
        """
        response = requests.get(
            url=self.api.base_url + "/Inverter/{}/kVa".format(self.id),
            headers={"Authorization": "Bearer " + self.api.access_token},
            params={"elevation": elevation, "temperature": temperature, "useCoolingTemp": use_cooling_temp}
        )
        
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)
        
        return response;

    @handle_refused_connection
    @handle_error_response
    def get_inverter_list(self):
        """
        :return: a list of all inverter to which a user has access.
        """
        return requests.get(
            url=self.api.base_url + "/Inverter",
            headers={"Authorization": "Bearer " + self.api.access_token},
           )



import requests
import json

from plantpredict.plant_predict_entity import PlantPredictEntity
from plantpredict.utilities import convert_json, snake_to_camel, camel_to_snake
from plantpredict.error_handlers import handle_refused_connection, handle_error_response, APIError
from plantpredict.enumerations import PredictionStatusEnum, EntityTypeEnum


class Prediction(PlantPredictEntity):
    """
    The :py:mod:`plantpredict.Prediction` entity models a single energy prediction within a
    :py:mod:`plantpredict.Project`.
    """
    def create(self, use_closest_ashrae_station=True, error_spa_var=2.0, error_model_acc=2.9, error_int_ann_var=3.0,
               error_sens_acc=5.0, error_mon_acc=2.0, year_repeater=1, status=PredictionStatusEnum.DRAFT_PRIVATE):
        """
        **POST** */Project/ :py:attr:`project_id` /Prediction*

        Creates a new :py:mod:`plantpredict.Prediction` entity in the PlantPredict database using the attributes
        assigned to the local object instance. Automatically assigns the resulting :py:attr:`id` to the local object
        instance. See the minimum required attributes (below) necessary to successfully create a new
        :py:mod:`plantpredict.Prediction`. Note that the full scope of attributes is not limited to the minimum
        required set. **Important Note:** the minimum required attributes necessary to create a
        :py:mod:`plantpredict.Prediction` is not sufficient to successfully call :py:meth:`plantpredict.Prediction.run`.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes for successful Prediction creation
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    name; str; Name of prediction
                    project_id; int; ID of project within which to contain the prediction
                    year_repeater; int; Must be between :py:data:`1` and :py:data:`50` - unitless.

        .. container:: toggle

            .. container:: header

                **Example Code**

            .. container:: example_code

                First, import the plantpredict library and receive an authentication api.access_token in your
                Python session, as shown in Step 3 of :ref:`authentication_oauth2`. Then instantiate a local Prediction.
                object.

                .. code-block:: python

                    module_to_create = plantpredict.Prediction()

                Populate the Prediction's require attributes by either directly assigning them...

                .. code-block:: python

                    from plantpredict.enumerations import PredictionStatusEnum

                    prediction_to_create.name = "Test Prediction"
                    prediction_to_create.project_id = 1000
                    prediction_to_create.status = PredictionStatusEnum.DRAFT_SHARED
                    prediction_to_create.year_repeater = 1

                ...OR via dictionary assignment.

                .. code-block:: python

                    prediction_to_create.__dict__ = {
                        "name": "Test Prediction",
                        "model": "Test Module",
                        "status": PredictionStatusEnum.DRAFT_SHARED,
                        "year_repeater": 1,
                    }

                Create a new prediction in the PlantPredict database, and observe that the Module now has a unique
                database identifier.

                .. code-block:: python

                    prediction_to_create.create()

                    print(prediction_to_create.id)

        :return: A dictionary containing the prediction id.
        :rtype: dict
        """

        self.create_url_suffix = "/Project/{}/Prediction".format(self.project_id)

        self.error_spa_var = error_spa_var
        self.error_model_acc = error_model_acc
        self.error_int_ann_var = error_int_ann_var
        self.error_sens_acc = error_sens_acc
        self.error_mon_acc = error_mon_acc
        self.year_repeater = year_repeater
        self.status = status

        if use_closest_ashrae_station:
            self._assign_plant_design_temperature_with_closest_ashrae_station()

        return super(Prediction, self).create()

    def _assign_plant_design_temperature_with_closest_ashrae_station(self):
        """
        Assigns the plant design temperatures by using the closest ASHRAE station (based on the associated project's
        latitude and longitude).
        """
        project = self.api.project(id=self.project_id)
        project.get()
        ashrae = self.api.ashrae(latitude=project.latitude, longitude=project.longitude)

        ashrae.get_closest_station()

        # set relevant attributes from ASHRAE to Prediction
        self.ashrae_station = ashrae.station_name
        self.cool_996 = ashrae.cool_996
        self.max_50_year = ashrae.max_50_year
        self.min_50_year = ashrae.min_50_year

    def delete(self):
        """HTTP Request: DELETE /Project/{ProjectId}/Prediction/{Id}

        Deletes an existing Prediction entity in PlantPredict. The local instance of the Project entity must have
        attribute self.id identical to the prediction id of the Prediction to be deleted.

        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """
        self.delete_url_suffix = "/Project/{}/Prediction/{}".format(self.project_id, self.id)

        return super(Prediction, self).delete()

    def get(self, id=None, project_id=None):
        """HTTP Request: GET /Project/{ProjectId}/Prediction/{Id}

        Retrieves an existing Prediction entity in PlantPredict and automatically assigns all of its attributes to the
        local Prediction object instance. The local instance of the Prediction entity must have attribute self.id
        identical to the prediction id of the Prediction to be retrieved.

        :return: A dictionary containing all of the retrieved Prediction attributes.
        :rtype: dict

        """
        self.id = id if id is not None else self.id
        self.project_id = project_id if project_id is not None else self.project_id

        self.get_url_suffix = "/Project/{}/Prediction/{}".format(self.project_id, self.id)

        return super(Prediction, self).get()

    def update(self):
        """HTTP Request: PUT /Project/{ProjectId}/Prediction

        Updates an existing Prediction entity in PlantPredict using the full attributes of the local Prediction
        instance. Calling this method is most commonly preceded by instantiating a local instance of Prediction with a
        specified prediction id, calling the Prediction.get() method, and changing any attributes locally.

        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """

        self.update_url_suffix = "/Project/{}/Prediction".format(self.project_id)

        return super(Prediction, self).update()

    @handle_refused_connection
    def _wait_for_prediction(self):
        is_complete = False
        while not is_complete:
            self.get()
            if self.processing_status == 3:
                is_complete = True

    @handle_refused_connection
    @handle_error_response
    def run(self, export_options=None):
        """
        POST /Project/{ProjectId}/Prediction/{PredictionId}/Run

        Runs the Prediction and waits for simulation to complete. The input variable "export_options" should take the

        :param export_options: Contains options for exporting
        :return:
        """
        response = requests.post(
            url=self.api.base_url + "/Project/{}/Prediction/{}/Run".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=convert_json(export_options, snake_to_camel) if export_options else None
        )
        # TODO why didn't this return an error? it only returned when I stopped the script

        # observes task queue to wait for prediction run to complete
        self._wait_for_prediction()

        return response

    @handle_refused_connection
    @handle_error_response
    def get_results_summary(self, negate_losses=False):
        """GET /Project/{ProjectId}/Prediction/{Id}/ResultSummary"""

        response = requests.get(
            url=self.api.base_url + "/Project/{}/Prediction/{}/ResultSummary".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)

        if negate_losses:
            results = json.loads(response.content)
            for year in results['years']:
                for factor in year['monthlyFactors']:
                    factor['soilingLoss'] *= -1
                    if factor['spectralShift'] is not None:
                        factor['spectralShift'] *= -1

                for ttl in year['transformerTransmissionLineLoss']:
                    ttl['loss'] *= -1

                year['transpositionOnPlane'] *= -1
                year['farShadingLoss'] *= -1
                year['nearShadingLoss'] *= -1
                year['elecShadingLoss'] *= -1
                year['soilingLoss'] *= -1
                year['iamFactorLoss'] *= -1
                year['spectralLoss'] *= -1
                year['moduleIrradianceLoss'] *= -1
                year['moduleTemperatureLoss'] *= -1
                year['moduleQualityLoss'] *= -1
                year['lidLoss'] *= -1
                year['moduleMismatchLoss'] *= -1
                year['moduleBackMismatchLoss'] *= -1
                year['biFacialityLoss'] *= -1
                year['structureShadingLoss'] *= -1
                year['backsideIrradiance'] *= -1
                year['dcWiringLoss'] *= -1
                year['dcHealthLoss'] *= -1
                year['inverterEfficiencyLoss'] *= -1
                year['inverterLimitationLoss'] *= -1
                year['degradationLoss'] *= -1
                year['leTIDLoss'] *= -1
                year['inverterCoolingLoss'] *= -1
                year['trackerMotorLoss'] *= -1
                year['dataAcquisitionAuxLoss'] *= -1
                year['mvTransformersLoss'] *= -1
                year['acCollectionLinesLoss'] *= -1
                year['availabilityLoss'] *= -1
                year['lgiaLimitationLoss'] *= -1

            # convert_json(response.json(), camel_to_snake)
            return convert_json(results, camel_to_snake)

        return response

    @handle_refused_connection
    @handle_error_response
    def get_results_details(self):
        """GET /Project/{ProjectId}/Prediction/{Id}/ResultDetails"""

        response = requests.get(
            url=self.api.base_url + "/Project/{}/Prediction/{}/ResultDetails".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)

        return response

    @handle_refused_connection
    @handle_error_response
    def get_nodal_data(self, params=None):
        """GET /Project/{ProjectId}/Prediction/{Id}/NodalJson"""

        response = requests.get(
            url=self.api.base_url + "/Project/{}/Prediction/{}/NodalJson".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token},
            params=convert_json(params, snake_to_camel) if params else {}
        )
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)
        return response

    @handle_refused_connection
    @handle_error_response
    def clone(self, new_prediction_name):
        """

        :param new_prediction_name:
        :return:
        """
        new_prediction = self.api.prediction()
        self.get()
        original_prediction_id = self.id

        new_prediction.__dict__ = self.__dict__
        # initialize necessary fields
        new_prediction.__dict__.pop('id', None)
        new_prediction.__dict__.pop('created_date', None)
        new_prediction.__dict__.pop('last_modified', None)
        new_prediction.__dict__.pop('last_modified_by', None)
        new_prediction.__dict__.pop('last_modified_by_id', None)
        new_prediction.__dict__.pop('project', None)
        new_prediction.__dict__.pop('powerplant_id', None)
        new_prediction.__dict__.pop('powerplant', None)

        new_prediction.name = new_prediction_name
        new_prediction.create()
        new_prediction_id = new_prediction.id

        # clone powerplant and attach to new prediction
        new_powerplant = self.api.powerplant()
        powerplant = self.api.powerplant(project_id=self.project_id, prediction_id=original_prediction_id)
        powerplant.get()
        new_powerplant.__dict__ = powerplant.__dict__
        new_powerplant.prediction_id = new_prediction_id
        new_powerplant.__dict__.pop('id', None)

        # initialize necessary fields
        for block in new_powerplant.blocks:
            block.pop('id', None)
            for array in block['arrays']:
                array.pop('id', None)
                for inverter in array['inverters']:
                    inverter.pop('id', None)
                    for dc_field in inverter['dc_fields']:
                        dc_field.pop('id', None)

        new_powerplant.create()

        self.id = original_prediction_id
        self.get()

        return new_prediction_id

    @handle_refused_connection
    @handle_error_response
    def change_status(self, new_status, note=""):
        """
        Change the status (and resulting sharing/privacy settings) of a prediction (ex. from py:attr:`DRAFT_PRIVATE` to
        py:attr:`DRAFT-SHARED`.

        :param int new_status: Enumeration representing status to change prediction to. See (or import)
                               :py:class:`plantpredict.enumerations.PredictionStatusEnum`.
        :param str note: Description of reason for change.
        :return:
        """
        return requests.post(
            url=self.api.base_url + "/Project/{}/Prediction/Status".format(self.project_id),
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=[{
                "name": self.name,
                "id": self.id,
                "type": EntityTypeEnum.PREDICTION,
                "status": new_status,
                "note": note
            }]
        )
    @handle_refused_connection
    @handle_error_response
    def get_time_series_data(self):
        """
        GET /Project/{ProjectId}/Prediction/{PredictionId}/TimeSeriesData

        :return:
        """
        return requests.get(
            url=self.api.base_url + "/Project/{}/Prediction/{}/TimeSeriesData".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

    @handle_refused_connection
    @handle_error_response
    def get_time_series_details(self, time_series_id):
        """
        GET /Project/{ProjectId}/Prediction/{PredictionId}/TimeSeriesData/{TimeSeriesId}

        :param time_series_id:
        :return:
        """
        request = requests.get(
            url=self.api.base_url + "/Project/{}/Prediction/{}/TimeSeriesData/{}/Details".format(self.project_id, self.id, time_series_id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

        return json.loads(request.content)
    @handle_refused_connection
    @handle_error_response
    def add_time_series_json(self, time_series_json):
        """
        POST /Project/{ProjectId}/Prediction/{PredictionId}/TimeSeriesData

        :param time_series_json:
        :return:
        """
        return requests.post(
            url=self.api.base_url + "/Project/{}/Prediction/{}/TimeSeriesData".format(self.project_id, self.id),
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=time_series_json
        )

    @handle_refused_connection
    @handle_error_response
    def delete_time_series_data(self, time_series_id):
        """
        DELETE /Project/{ProjectId}/Prediction/{PredictionId}/TimeSeriesData/{TimeSeriesId}

        :param time_series_id:
        :return:
        """
        return requests.delete(
            url=self.api.base_url + "/Project/{}/Prediction/{}/TimeSeriesData/{}".format(self.project_id, self.id, time_series_id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

    def __init__(self, api, id=None, project_id=None, name=None):
        if id:
            self.id = id
        self.project_id = project_id
        self.name = name

        self.status = None
        self.year_repeater = None

        self.error_spa_var = None
        self.error_model_acc = None
        self.error_int_ann_var = None
        self.error_sens_acc = None
        self.error_mon_acc = None

        super(Prediction, self).__init__(api)



import mock
import json
import unittest

from plantpredict.project import Project
from tests import plantpredict_unit_test_case, mocked_requests


class TestProject(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.create')
    def test_create(self, mocked_create):
        self._make_mocked_api()
        project = Project(api=self.mocked_api)

        project.create()
        self.assertEqual(project.create_url_suffix, "/Project")
        self.assertTrue(mocked_create.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.delete')
    def test_delete(self, mocked_delete):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, id=710)

        project.delete()
        self.assertEqual(project.delete_url_suffix, "/Project/710")
        self.assertTrue(mocked_delete.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.get')
    def test_get(self, mocked_get):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, id=710)

        project.get()
        self.assertEqual(project.get_url_suffix, "/Project/710")
        self.assertTrue(mocked_get.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.update')
    def test_update(self, mocked_update):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, id=710)

        project.update()
        self.assertEqual(project.update_url_suffix, "/Project")
        self.assertTrue(mocked_update.called)

    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_all_predictions(self):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, id=710)

        response = project.get_all_predictions()
        self.assertEqual(response.json(), [
            {"project_id": 1, "name": "Project 1"},
            {"project_id": 2, "name": "Project 2"},
            {"project_id": 3, "name": "Project 3"}
        ])

    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_search(self):
        self._make_mocked_api()
        project = Project(api=self.mocked_api)

        response = project.search(latitude=33.0, longitude=-118.0, search_radius=2)
        self.assertEqual(response, [{"project_id": 1, "name": "Project 1"}])

    def test_assign_location_attributes(self):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, latitude=39.67, longitude=-105.21, name="Test Project")

        project.assign_location_attributes()
        self.assertEqual(project.country, "United States")
        self.assertEqual(project.country_code, "US")
        self.assertEqual(project.locality, "Morrison")
        self.assertEqual(project.region, "North America")
        self.assertEqual(project.state_province, "Colorado")
        self.assertEqual(project.state_province_code, "CO")

    def test_init_without_id(self):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, latitude=39.67, longitude=-105.21, name="Test Project")

        with self.assertRaises(AttributeError):
            print(project.id)
        self.assertEqual(project.api, self.mocked_api)
        self.assertEqual(project.latitude, 39.67)
        self.assertEqual(project.longitude, -105.21)
        self.assertEqual(project.name, "Test Project")
        self.assertIsNone(project.locality)
        self.assertIsNone(project.state_province_code)
        self.assertIsNone(project.state_province)
        self.assertIsNone(project.country_code)
        self.assertIsNone(project.country)
        self.assertIsNone(project.region)
        self.assertIsNone(project.elevation)
        self.assertIsNone(project.standard_offset_from_utc)

    def test_init_with_id(self):
        self._make_mocked_api()
        project = Project(api=self.mocked_api, latitude=39.67, longitude=-105.21, name="Test Project", id=101)

        self.assertEqual(project.id, 101)
        self.assertEqual(project.api, self.mocked_api)
        self.assertEqual(project.latitude, 39.67)
        self.assertEqual(project.longitude, -105.21)
        self.assertEqual(project.name, "Test Project")
        self.assertIsNone(project.locality)
        self.assertIsNone(project.state_province_code)
        self.assertIsNone(project.state_province)
        self.assertIsNone(project.country_code)
        self.assertIsNone(project.country)
        self.assertIsNone(project.region)
        self.assertIsNone(project.elevation)
        self.assertIsNone(project.standard_offset_from_utc)


if __name__ == '__main__':
    unittest.main()



import re


def decorate_all_methods(decorator):
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr != '__init__':
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


def camel_to_snake(key):
    camel_pat = re.compile(r'([A-Z])')
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), key)


def snake_to_camel(key):
    under_pat = re.compile(r'_([a-z])')
    return under_pat.sub(lambda x: x.group(1).upper(), key)


MANUAL_KEY_FIXES = {
    "camel_to_snake": {
        "k_m_l": "kml",
        "s_t_c": "stc",
        "STC": "stc",
        "i_a_m": "iam",
        "power_plant": "powerplant",
        "p_o_a": "poa",
        "back_tracking": "backtracking",
        "back_side": "backside",
        "k_v_a": "kva",
        "transformerkva_rating": "transformer_kva_rating",
        "inverterkva": "inverter_kva",
        "usek_v_a": "use_kva_",
        "m_v": "mv",
        "e_s_s": "ess",
        "k_w": "_kw",
        "m_p_p": "mpp",
        "d_c": "dc",
        "p_c_s": "pcs",
        "uiamd": "uiam_d",
        "uiamg": "uiam_g",
        "ashraeiam": "ashrae_iam",
        "stcmpp": "stc_mpp",
        "a_c": "ac",
        "u_t_c": "utc",
        "g_h_i": "ghi",
        "d_h_i": "dhi",
        "d_n_i": "dni",
        "poa_i": "poai",
        "g_c_r": "gcr",
        "bi_faciality": "bifaciality",
        "poairradinace": "global_poa_irradiance",
        "g_teff": "gt_eff",
        "time_stamp": "timestamp",
        "i_v": "iv",
        "po_a_insolation": "poa_insolation",
        "light_generatedcurrent": "light_generated_current",
        "bo_s": "bos",
        "sandiaconductive": "sandia_conductive",
        "sandiaconvective": "sandia_convective",
        "l_i_d": "lid",
        "at25": "at_25",
        "l_g_i_a": "lgia",
        "cool996": "cool_996",
        "heat996": "heat_996",
        "max50_year": "max_50_year",
        "min50_year": "min_50_year",
        "p_q": "pq",
        "kvacurves": "kva_curves",
        "k_va": "kva",
    },
    "snake_to_camel": {
        "powerplant": "powerPlant",
        "backtracking": "backTracking",
        "backsideMismatch": "backSideMismatch"
    }
}


def convert_json(d, convert_function):
    """
    Convert a nested dictionary from one convention to another. Prepares payload for http request.
    Args:
        d (dict): dictionary (nested or not) to be converted.
        convert_function (func): function that takes the string in one convention and returns it in the other one.
    Returns:
        Dictionary with the new keys.

    """
    # "api" object is not serializable, so remove it from http request
    dict_copy = d.copy()
    dict_copy.pop("api", None)

    new = {}
    for k, v in dict_copy.items():
        new_v = v
        if isinstance(v, dict):
            new_v = convert_json(v, convert_function)
        elif isinstance(v, list):
            new_v = list()
            for x in v:
                if isinstance(x, dict):
                    new_v.append(convert_json(x, convert_function))

        new_key = convert_function(k)

        # manual fixes
        for key, val in MANUAL_KEY_FIXES[convert_function.__name__].items():
            if key in new_key:
                if not (key == "d_c" and new_key == "light_generated_current"):       # edge case
                    new_key = new_key.replace(key, val)

        # this removes the underscore given to a snake case when the first character in the camel case is capital
        new_key = new_key[1:] if new_key[0] == "_" else new_key

        new[new_key] = new_v

    return new


def convert_json_list(l, convert_function):
    new_list = []
    for d in l:
        # "api" object is not serializable, so remove it from http request
        dict_copy = d.copy()
        dict_copy.pop("api", None)

        new = {}
        for k, v in dict_copy.items():
            new_v = v
            if isinstance(v, dict):
                new_v = convert_json(v, convert_function)
            elif isinstance(v, list):
                new_v = list()
                for x in v:
                    if isinstance(x, dict):
                        new_v.append(convert_json(x, convert_function))

            new_key = convert_function(k)

            # manual fixes
            for key, val in MANUAL_KEY_FIXES[convert_function.__name__].items():
                if key in new_key:
                    if not (key == "d_c" and new_key == "light_generated_current"):       # edge case
                        new_key = new_key.replace(key, val)

            # this removes the underscore given to a snake case when the first character in the camel case is capital
            new_key = new_key[1:] if new_key[0] == "_" else new_key

            new[new_key] = new_v

        new_list.append(new)

    return new_list



import os
from setuptools import setup

__currdir__ = os.getcwd()
__readme__ = os.path.join(__currdir__, 'README.md')


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name='plantpredict',
    version=get_version("plantpredict/__init__.py"),
    description='Python SDK for PlantPredict (https://ui.plantpredict.terabase.energy).',
    url='https://github.com/plantpredict/python-sdk',
    author='Kurt Rhee & Jese Milam.',
    author_email='support@plantpredic.com',
    license='LICENSE.txt',
    long_description=open(__readme__).read(),
    packages=['plantpredict'],
    python_requires='>=3.10, <4',
    install_requires=[
        'requests',
        'pandas',
        'six',
        'mock',
        'xlrd',
        'openpyxl'
    ]
)
